/*
 * html.c
 *
 * Copyright (c) 1994, 1995, John D. Kilburg <john@cs.unlv.edu>
 * See the file COPYRIGHT for more information.
 */
#include <stdio.h>
#include <ctype.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include "common.h"
#include "html.h"

static MLAttribute *hs_parse_tag _ArgProto((MLState *, char *, size_t));
static void hs_create_part _ArgProto((MLState *, int,
				      size_t, size_t,
				      char *, size_t,
				      MLAttribute *, bool));
static void hs_create_text _ArgProto((MLState *, size_t, size_t));
static void hs_create_tag _ArgProto((MLState *, size_t, size_t));
static char *hs_condense _ArgProto((MLState *, char *, size_t, size_t *));
static MLAttribute *hs_create_attribute _ArgProto((MLState *, char *, size_t));

MLTagName tname[] =
{
  { "a", TAG_A },
  { "address", TAG_ADDRESS },
  { "abbrev", TAG_ABBREV },
  { "above", TAG_ABOVE },
  { "acronym", TAG_ACRONYM },
  { "address", TAG_ADDRESS },
  { "arow", TAG_AROW },
  { "array", TAG_ARRAY },
  { "atop", TAG_ATOP },
  { "au", TAG_AU },
  { "b", TAG_B },
  { "base", TAG_BASE },
  { "below", TAG_BELOW },
  { "big", TAG_BIG },
  { "blockquote", TAG_BLOCKQUOTE },
  { "bq", TAG_BLOCKQUOTE },
  { "body", TAG_BODY },
  { "box", TAG_BOX },
  { "bq", TAG_BQ },
  { "br", TAG_BR },
  { "caption", TAG_CAPTION },
  { "cite", TAG_CITE },
  { "code", TAG_CODE },
  { "dd", TAG_DD },
  { "dfn", TAG_DFN },
  { "dir", TAG_DIR },
  { "dl", TAG_DL },
  { "dt", TAG_DT },
  { "em", TAG_EM },
  { "fig", TAG_FIG },
  { "fn", TAG_FN },
  { "form", TAG_FORM },
  { "h1", TAG_H1 },
  { "h2", TAG_H2 },
  { "h3", TAG_H3 },
  { "h4", TAG_H4 },
  { "h5", TAG_H5 },
  { "h6", TAG_H6 },
  { "head", TAG_HEAD },
  { "hr", TAG_HR },
  { "html", TAG_HTML },
  { "i", TAG_I },
  { "img", TAG_IMG },
  { "input", TAG_INPUT },
  { "isindex", TAG_ISINDEX },
  { "item", TAG_ITEM },
  { "kbd", TAG_KBD },
  { "lang", TAG_LANG },
  { "lh", TAG_LH },
  { "li", TAG_LI },
  { "link", TAG_LINK },
  { "listing", TAG_LISTING },
  { "math", TAG_MATH },
  { "menu", TAG_MENU },
  { "meta", TAG_META },
  { "nextid", TAG_NEXTID },
  { "ol", TAG_OL },
  { "option", TAG_OPTION },
  { "over", TAG_OVER },
  { "overlay", TAG_OVERLAY },
  { "p", TAG_P },
  { "person", TAG_PERSON },
  { "plaintext", TAG_PLAINTEXT },
  { "pre", TAG_PRE },
  { "q", TAG_Q },
  { "right", TAG_RIGHT },
  { "root", TAG_ROOT },
  { "s", TAG_S },
  { "samp", TAG_SAMP },
  { "select", TAG_SELECT },
  { "small", TAG_SMALL },
  { "strong", TAG_STRONG },
  { "style", TAG_STYLE },
  { "styles", TAG_STYLES },
  { "sub", TAG_SUB },
  { "sup", TAG_SUP },
  { "tab", TAG_TAB },
  { "table", TAG_TABLE },
  { "td", TAG_TD },
  { "textarea", TAG_TEXTAREA },
  { "th", TAG_TH },
  { "title", TAG_TITLE },
  { "tr", TAG_TR },
  { "tt", TAG_TT },
  { "u", TAG_U },
  { "ul", TAG_UL },
  { "var", TAG_VAR },
  { "xmp", TAG_XMP },
  { NULL, 0 },
};

/*
 * hs_create_attribute
 */
static MLAttribute *
hs_create_attribute(hs, text, len)
MLState *hs;
char *text;
size_t len;
{
  MLAttribute *ta;

  ta = (MLAttribute *)MPCGet(hs->mp, sizeof(MLAttribute));

  ta->name = (char *)MemPoolGet(hs->mp, len + 1);
  strncpy(ta->name, text, len);
  ta->name[len] = '\0';

  return(ta);
}

struct entity
{
  char *text;
  int len;
  unsigned char c;
} clist[] =
{
  { "amp", 3, '&' },
  { "lt", 2, '<' },
  { "gt", 2, '>' },
  { "copy", 4, 169 },
  { "AElig", 5, 198 },
  { "Aacute", 6, 193 },
  { "Acirc", 5, 194 },
  { "Agrave", 6, 192 },
  { "Aring", 5, 197 },
  { "Atilde", 6, 195 },
  { "Auml", 4, 196 },
  { "Ccedil", 6, 199 },
  { "ETH", 3, 208 },
  { "Eacute", 6, 201 },
  { "Ecirc", 5, 202 },
  { "Egrave", 6, 200 },
  { "Euml", 4, 203 },
  { "Iacute", 6, 205 },
  { "Icirc", 5, 206 },
  { "Igrave", 6, 204 },
  { "Iuml", 4, 207 },
  { "Ntilde", 6, 209 },
  { "Oacute", 6, 211 },
  { "Ocirc", 5, 212 },
  { "Ograve", 6, 210 },
  { "Oslash", 6, 216 },
  { "Otilde", 6, 213 },
  { "Ouml", 4, 214 },
  { "THORN", 5, 222 },
  { "Uacute", 6, 218 },
  { "Ucirc", 5, 219 },
  { "Ugrave", 6, 217 },
  { "Uuml", 4, 220 },
  { "Yacute", 6, 221 },
  { "aacute", 6, 225 },
  { "acirc", 5, 226 },
  { "aelig", 5, 230 },
  { "agrave", 6, 224 },
  { "aring", 5, 229 },
  { "atilde", 6, 227 },
  { "auml", 4, 228 },
  { "ccedil", 6, 231 },
  { "eacute", 6, 233 },
  { "ecirc", 5, 234 },
  { "egrave", 6, 232 },
  { "eth", 3, 240 },
  { "euml", 4, 235 },
  { "iacute", 6, 237 },
  { "icirc", 5, 238 },
  { "igrave", 6, 236 },
  { "iuml", 4, 239 },
  { "ntilde", 6, 241 },
  { "oacute", 6, 243 },
  { "ocirc", 5, 244 },
  { "ograve", 6, 242 },
  { "oslash", 6, 248 },
  { "otilde", 6, 245 },
  { "ouml", 4, 246 },
  { "szlig", 5, 223 },
  { "thorn", 5, 254 },
  { "uacute", 6, 250 },
  { "ucirc", 5, 251 },
  { "ugrave", 6, 249 },
  { "uuml", 4, 252 },
  { "yacute", 6, 253 },
  { "yuml", 4, 255 },
  { "reg", 3, 174 },
  { "comma", 5, 44 },
  { "colon", 5, 58 },
  { "quot", 4, '\"' },
  { "nbsp", 4, ' ' },
  { NULL, 0, '\0' },
};

/*
 * hs_condense
 */
static char *
hs_condense(hs, text, len, newlen)
MLState *hs;
char *text;
size_t len;
size_t *newlen;
{
  char *cp, *lastcp;
  size_t i;
  char *b, *bcp;
  char *x;

  /*
   * allocate a buffer for the new text.  the new text will be smaller than
   * the incoming text
   */
  b = (char *)MemPoolGet(hs->mp, len + 1);
  bcp = b;

  for (cp = text, lastcp = text + len; cp < lastcp; cp++)
  {
    /*
     * Found entity and there is at least one more character?
     */
    if (*cp == '&')
    {
      for (x = cp; x < lastcp && *x != ';' && !isspace(*x); x++)
	  ;
      if (x < lastcp)                              /* found end of entity ? */
      {
	if (*x == '\r') x++;                       /* this may die */
	if (*(cp + 1) == '#')                      /* number given ? */
	{
	  *bcp = (char)atoi(cp + 2);               /* get the character */
	  bcp++;
	  cp = x;
	}
	else
	{
	  char *crap;

	  /* Search in the table of entity names */
	  for (i = 0, crap = cp + 1; clist[i].text != NULL; i++)
	  {
	    if (strncmp(clist[i].text, crap, clist[i].len) == 0)
	    {
	      *bcp = clist[i].c;
	      bcp++;
	      cp = x;
	      break;
	    }
	  }
	}
      }
      if (cp < x)
      {
	*bcp = *cp;                 
	bcp++;
      }
    }
    else
    {
      *bcp = *cp;                                 /* not part of entity */
      bcp++;
    }
  }

  *newlen = bcp - b;
  *bcp = '\0';

  return(b);
}

/*
 * hs_parse_tag
 */
MLAttribute *
hs_parse_tag(hs, text, len)
MLState *hs;
char *text;
size_t len;
{
  char *cp, *start, *lastcp;
  MLAttribute *tahead, *tatail, *ta;
  size_t newlen;

  tahead = NULL;
  tatail = NULL;
  lastcp = text + len;
  for (cp = text; cp < lastcp; )
  {
    /* Eat leading spaces */
    for (; cp < lastcp && isspace(*cp); cp++)
	;
    if (cp == lastcp) break;

    /* Look for value */
    for (start = cp; cp < lastcp && !isspace(*cp) && *cp != '='; cp++)
	;

    ta = hs_create_attribute(hs, start, cp - start);

    if (isspace(*cp))
    {
      for (; cp < lastcp && isspace(*cp); cp++)
	  ;
    }

    /* Found value */
    if (cp < lastcp && *cp == '=')
    {
      cp++; /* past '=' */

      /* Eat leading spaces */
      for (; cp < lastcp && isspace(*cp); cp++)
	  ;
      if (cp < lastcp)
      {
	/* Quoted value */
	if (*cp == '"')
	{
	  cp++; /* past '"' */

	  for (start = cp; cp < lastcp && *cp != '"'; cp++)
	      ;
	}
	else if (*cp == '\'') /* single quoted value */
	{
	  cp++; /* past ' */

	  for (start = cp; cp < lastcp && *cp != '\''; cp++)
	      ;
	}
	else /* Unquoted value */
	{
	  for (start = cp; cp < lastcp && !isspace(*cp); cp++)
	      ;
	}

	ta->value = hs_condense(hs, start, cp - start, &newlen);
      }
    }

    if (tatail == NULL) tahead = ta;
    else tatail->next = ta;
    tatail = ta;
  }

  return(tahead);
}


/*
 * hs_create_part
 */
static void
hs_create_part(hs, tagid, ooff, olen, text, len, ta, end)
MLState *hs;
int tagid;
size_t ooff;
size_t olen;
char *text;
size_t len;
MLAttribute *ta;
bool end;
{
  MLTag *p;

  p = (MLTag *)MPCGet(hs->mp, sizeof(MLTag));
  p->id = tagid;
  p->otext = (char *)MemPoolGet(hs->mp, olen);
  strncpy(p->otext, hs->data + ooff, olen);
  p->olen = olen;
  p->text = text;
  p->len = len;
  p->next = NULL;
  p->qnext = NULL;
  p->end = end;
  p->ta = ta;

  if (hs->ptail != NULL) hs->ptail->next = p;
  else hs->phead = p;
  hs->ptail = p;

  (*(hs->func))(hs->closure, p);

  return;
}

/*
 * hs_create_text
 */
static void
hs_create_text(hs, toff, len)
MLState *hs;
size_t toff;
size_t len;
{
  char *cp, *text;
  size_t newlen;
  size_t i;

  /* Check for an entity */
  text = hs->data + toff;
  for (cp = text, i = 0; i < len && *cp != '&'; cp++, i++)
      ;

  if (i == len) hs_create_part(hs, TAG_TEXT, toff, len, NULL, 0, NULL, false);
  else
  {
    /* found an entity...condense */
    cp = hs_condense(hs, text, len, &newlen);
    hs_create_part(hs, TAG_TEXT, toff, len, cp, newlen, NULL, false);
  }

  return;
}

/*
 * hs_create_tag
 */
static void
hs_create_tag(hs, toff, len)
MLState *hs;
size_t toff;
size_t len;
{
  char *name;
  bool end;
  size_t i;
  MLAttribute *ta;

  if ((ta = hs_parse_tag(hs, hs->data + toff + 1, len - 2)) == NULL) return;

  if (*(ta->name) == '/')
  {
    name = ta->name + 1;
    end = true;
  }
  else
  {
    name = ta->name;
    end = false;
  }

  for (i = 0; tname[i].name != NULL; i++)
  {
    if (strcasecmp(tname[i].name, name) == 0)
    {
      hs_create_part(hs, tname[i].id, toff, len, NULL, 0, ta, end);
      return;
    }
  }

  hs_create_part(hs, TAG_UNKNOWN, toff, len, NULL, 0, ta, end);

  return;
}

/*
 * MLInit
 */
MLState *
MLInit(mp, broken, func, closure)
MemPool mp;
bool broken;
void (*func)();
void *closure;
{
  MLState *hs;

  hs = (MLState *)MPCGet(mp, sizeof(MLState));
  hs->mp = mp;
  hs->broken = broken;
  hs->func = func;
  hs->closure = closure;
  hs->context = HTML_IN_TAG;
  
  return(hs);
}

/*
 * MLDestroy
 */
void
MLDestroy(hs)
MLState *hs;
{
  return;
}

#define HA(x) (*(data + x->off))
#define INCOMMENT(x) (x->context == HTML_IN_COMMENT)
#define INQUOTES(x) (x->inquotes == 1)
#define INTAG(x) (x->context == HTML_IN_TAG)
#define INTEXT(x) (x->context == HTML_IN_TEXT)

/*
 * MLAddData
 */
void
MLAddData(hs, data, len)
MLState *hs;
char *data;
size_t len;
{
  char *cp;
  int old_context;

  hs->data = data;
  hs->len = len;

  if (hs->off == 0 && len > 0)
  {
    if (*data != '<') hs->context = HTML_IN_TEXT;
  }

  for (; hs->off < len; hs->off++)
  {
    cp = data + hs->off;
    if (*cp == '<' && !INCOMMENT(hs) && !INQUOTES(hs))
    {
      old_context = hs->context;

      if (hs->off + 1 == len) return;
      else if (*(cp + 1) == '!')
      {
	if (hs->off + 2 == len) return;
	else if (*(cp + 2) == '-')
	{
	  if (hs->off + 3 == len) return;
	  else if (*(cp + 3) == '-') hs->context = HTML_IN_COMMENT;
	  else hs->context = HTML_IN_TAG;
	}
	else hs->context = HTML_IN_TAG;
      }
      else hs->context = HTML_IN_TAG;

      if (old_context == HTML_IN_TEXT)
      {
	hs_create_text(hs, hs->coff, hs->off - hs->coff);
      }
      hs->coff = hs->off;
    }
    else if (*cp == '>' && !INQUOTES(hs) && !INCOMMENT(hs))
    {
      if (INTAG(hs))
      {
	hs_create_tag(hs, hs->coff, hs->off - hs->coff + 1);
	hs->coff = hs->off + 1;
	hs->context = HTML_IN_TEXT;
      }
    }
    else if (INCOMMENT(hs) && !INQUOTES(hs) && *cp == '-')
    {
      if ((hs->off + 1) == len) return;
      else if (*(cp + 1) == '-')
      {
	if ((hs->off + 2) == len) return;
	else if (*(cp + 2) == '>')
        {
	  hs_create_tag(hs, hs->coff, hs->off - hs->coff + 1);
	  hs->context = HTML_IN_TEXT;
	  hs->off += 2;
	  hs->coff = hs->off + 1;
	}
      }
    }
    else if (*cp == '"' && INTAG(hs))
    {
      if (!hs->broken) hs->inquotes = 1 - hs->inquotes;
    }
  }

  return;
}

/*
 * MLEndData
 */
void
MLEndData(hs, data, len)
MLState *hs;
char *data;
size_t len;
{
  MLAddData(hs, data, len);

  if (INTEXT(hs))
  {
    hs_create_text(hs, hs->coff, len - hs->coff);
  }
  else if (INTAG(hs) || INCOMMENT(hs))
  {
    hs_create_tag(hs, hs->coff, len - hs->coff);
  }
  
  hs_create_part(hs, TAG_EOF, 0, 0, NULL, 0, NULL, 0);
  
  return;
}

/*
 * MLFindAttribute
 */
char *
MLFindAttribute(p, name)
MLTag *p;
char *name;
{
  MLAttribute *ta;

  for (ta = p->ta; ta != NULL; ta = ta->next)
  {
    if (strcasecmp(ta->name, name) == 0)
    {
      return(ta->value == NULL ? "":ta->value);
    }
  }

  return(NULL);
}

/*
 * MLGetAlignID
 */
int
MLGetAlignID(s)
char *s;
{
  int align_id;

  if (strcasecmp(s, "left") == 0) align_id = ALIGN_LEFT;
  else if (strcasecmp(s, "center") == 0) align_id = ALIGN_CENTER;
  else if (strcasecmp(s, "right") == 0) align_id = ALIGN_RIGHT;
  else if (strcasecmp(s, "top") == 0) align_id = ALIGN_TOP;
  else if (strcasecmp(s, "bottom") == 0) align_id = ALIGN_BOTTOM;
  else align_id = ALIGN_JUSTIFY;

  return(align_id);
}

/*
 * MLGetTagText
 */
void
MLGetTagText(hs, p, original, text, len)
MLState *hs;
MLTag *p;
bool original;
char **text;
size_t *len;
{
  if (original || p->text == NULL)
  {
    *text = p->otext;
    *len = p->olen;
  }
  else
  {
    *text = p->text;
    *len = p->len;
  }

  return;
}
