/*
 * inline.c
 *
 * Copyright (c) 1995, John Kilburg <john@cs.unlv.edu>
 * See the file COPYRIGHT for details.
 */
#include <stdio.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include "layout.h"

typedef struct inline_state
{
  LayoutInfo *li;  /* need this always */
  bool added;      /* added yet */
  Frame frame;     /* embedded document ID */
} InlineState;

/*
 *
 * Private Functions
 *
 */
static void SetupInline _ArgProto((LayoutInfo *, RenderPart *, int, int));

/*
 * SetupInline
 */
static void
SetupInline(li, rp, line_y, line_height)
LayoutInfo *li;
RenderPart *rp;
int line_y;
int line_height;
{
  InlineState *ci = GetRenderState(rp);

  rp->y = line_y + line_height - rp->height;
  rp->renderable = True;

  FrameMap(ci->frame, rp->x, rp->y);

  return;
}

/*
 * DestroyInline
 */
static void
DestroyInline(ci)
InlineState *ci;
{
  if (ci->frame != NULL) FrameDestroy(ci->frame);

  return;
}

/*
 * RealizeCallback
 */
static void
RealizeCallback(rp, status, width, height)
RenderPart *rp;
int status;
unsigned int width, height;
{
  InlineState *ci = GetRenderState(rp);

  if (ci->added) return;
  ci->added = True;

  if (status == -1)
  {
    rp = CreateRenderPart(ci->li, NULL, NULL, NULL, NULL, NULL, 0);
    AddRenderPart(ci->li, rp, 0, 0);
  }
  else
  {
    AddRenderPart(ci->li, rp, width, height);
  }

  return;
}

/*
 * MouseCallback
 */
static int
InlineMouseCallback(closure, x, y, button)
void *closure;
int x, y;
int button;
{
  RenderPart *rp = (RenderPart *)closure;
  InlineState *ci = GetRenderState(rp);
  LayoutInfo *li = ci->li;
  URLParts *up, *lup;
  char *url;
  char *aurl;

  if (rp->anchorPart == NULL || rp->p == NULL) return(0);
  if ((url = MLFindAttribute(rp->anchorPart, "href")) == NULL) return(0);

  if (MLFindAttribute(rp->p, "ismap") != NULL)
  {
    aurl = (char *)MPGet(li->mp, strlen(url) + 200);
    sprintf (aurl, "%s?%d,%d", url, x, y);
    url = aurl;
  }
  else aurl = NULL;

  if ((lup = URLParse(li->mp, url)) != NULL)
  {
    if (li->bup != NULL) up = URLResolve(li->mp, lup, li->bup);
    else up = lup;
    
    /*
     * Careful with these two...may lose LayoutInfo before they
     * return.
     */
    if (button == 1) FrameLoad(li->frame, up);
    else if (button == 2) FrameDownload(li->frame, up);
  }
  
  return(0);
}

static int
InlineMotionCallback(closure, x, y, button)
void *closure;
int x, y;
int button;
{
  RenderPart *rp = (RenderPart *)closure;
  InlineState *ci = GetRenderState(rp);
  LayoutInfo *li = ci->li;
  char *url;
  URLParts *lup;
  MemPool mp;

  if (li->over == rp) return(0);
  if (rp->anchorPart == NULL || rp->p == NULL) return(0);
  if ((url = MLFindAttribute(rp->anchorPart, "href")) == NULL) return(0);

  mp = MPCreate();
  if ((lup = URLParse(mp, url)) != NULL)
  {
    URLParts *up;
    char *turl;
    
    if (li->bup != NULL) up = URLResolve(mp, lup, li->bup);
    else up = lup;
    
    turl = URLMakeString(mp, up, 0);
    FrameMessage(li->frame, turl);
  }
  MPDestroy(mp);

  li->over = rp;

  return(0);
}

/*
 *
 * Public functions
 *
 */

/*
 * HandleImg
 */
void
HandleImg(li, p, end)
LayoutInfo *li;
MLTag *p;
int end;
{
  char *url;
  URLParts *aup;
  URLParts *lup;
  InlineState *ci;
  RenderPart *rp;
  char *ws, *hs;
  RenderHooks orh;
  
  if (end != 0) return;
  if ((url = MLFindAttribute(p, "src")) == NULL) return;
  if ((lup = URLParse(li->mp, url)) == NULL) return;
  if (li->bup != NULL) aup = URLResolve(li->mp, lup, li->bup);
  else aup = lup;

  rp = CreateRenderPart(li, p,
			SetupInline, NULL, DestroyInline, NULL,
			sizeof(InlineState));
  ci = (InlineState *)GetRenderState(rp);
  ci->li = li;
  ci->added = False;

  li->delay = True;

  if ((ws = MLFindAttribute(p, "width")) != NULL &&
      (hs = MLFindAttribute(p, "height")) != NULL)
  {
    unsigned int width, height;

    width = (unsigned int)atoi(ws);
    height = (unsigned int)atoi(hs);
    ci->frame = FrameCreateSized(li->frame, width, height);
    RealizeCallback(rp, 0, width, height);
  }
  else
  {
    ci->frame = FrameCreateUnsized(li->frame, RealizeCallback, rp);
  }

  memset(&orh, 0, sizeof(RenderHooks));
  orh.mouse = InlineMouseCallback;
  orh.motion = InlineMotionCallback;
  FrameOverride(ci->frame, &orh, rp);
  FrameLoad(ci->frame, aup);

  return;
}
