/*
 * layout.c
 *
 * Copyright (c) 1995, John D. Kilburg <john@cs.unlv.edu>
 * See the file COPYRIGHT for details.
 */
#include <stdio.h>
#include <ctype.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include "layout.h"

static RenderLine *CreateRenderLine _ArgProto((LayoutInfo *));
static void CallTagHandler _ArgProto((LayoutInfo *, int, MLTag *));
static void DisplayLine _ArgProto((LayoutInfo *, RenderLine *));

/*
 * LayoutSpacify
 *
 * Turns all whitespace to spaces
 */
void
LayoutSpacify(s, len)
char *s;
size_t len;
{
  char *cp, *lastcp = s + len;

  for (cp = s; cp < lastcp; cp++)
  {
    if (isspace(*cp)) *cp = ' ';
  }

  return;
}

/*
 * RemainingLineWidth
 */
unsigned int
RemainingLineWidth(li)
LayoutInfo *li;
{
  RenderLine *rl;

  rl = li->rltail;

  return(rl->maxwidth - rl->width);
}

/*
 * GetWidthHeight
 */
int
GetWidthHeight(p, width, height)
MLTag *p;
unsigned int *width, *height;
{
  char *wstr, *hstr;

  wstr = MLFindAttribute(p, "width");
  hstr = MLFindAttribute(p, "height");
  if (wstr != NULL && hstr != NULL)
  {
    *width = (unsigned int)atoi(wstr);
    *height = (unsigned int)atoi(hstr);
    return(0);
  }

  return(-1);
}

void
AddLineBreak(li)
LayoutInfo *li;
{
  RenderLine *orl = li->rltail;

  CreateRenderLine(li);

  /* Paint the line after the new one is created */
  if (orl != NULL) DisplayLine(li, orl);

  return;
}

void
AddBlankLine(li)
LayoutInfo *li;
{
  AddLineBreak(li);
  AddLineBreak(li);

  return;
}

/*
 * InEnv
 */
bool
InEnv(li, tagid)
LayoutInfo *li;
int tagid;
{
  RenderEnv *re;

  for (re = li->restack; re != NULL; re = re->next)
  {
    if (re->tagid == tagid) return(true);
  }

  return(false);
}

/*
 * CreateRenderEnv
 */
RenderEnv *
CreateRenderEnv(li, tagid)
LayoutInfo *li;
int tagid;
{
  RenderEnv *re;

  re = (RenderEnv *)MPCGet(li->mp, sizeof(RenderEnv));
  re->tagid = tagid;

  if (li->restack != NULL)
  {
    RenderEnv *ce = li->restack;

    re->anchorPart = ce->anchorPart;
    re->namePart = ce->namePart;
    memcpy(&(re->fi), &(ce->fi), sizeof(LayoutFontInfo));
    re->align = ce->align;
    re->text_mode = ce->text_mode;
    re->x = ce->x;
    re->width = ce->width;
    re->wrap = ce->wrap;
  }
  else
  {
    memcpy(&(re->fi), &(li->cfi), sizeof(LayoutFontInfo));
    re->align = ALIGN_LEFT;
    re->text_mode = TAG_NOTHING;
    re->x = li->margin;
    re->width = li->width;
    re->wrap = true;
  }

  re->next = li->restack;
  li->restack = re;

  return(re);
}

/*
 * DestroyRenderEnv2
 */
void
DestroyRenderEnv2(li, tagid, call_self)
LayoutInfo *li;
int tagid;
bool call_self;
{
  RenderEnv *fe, *te;

  if (li->destroy_locked) return;
  li->destroy_locked = true;

  for (fe = li->restack; fe != NULL; fe = fe->next)
  {
    if (tagid == fe->tagid) break;
  }

  if (fe != NULL)
  {
    for (; li->restack != NULL && li->restack != fe; )
    {
      te = li->restack;
      CallTagHandler(li, te->tagid, NULL);
      li->restack = li->restack->next;
    }
    if (call_self) CallTagHandler(li, fe->tagid, NULL);
    li->restack = fe->next;
  }

  li->destroy_locked = false;

  return;
}

/*
 * DestroyRenderEnv
 */
void
DestroyRenderEnv(li, tagid)
LayoutInfo *li;
int tagid;
{
  DestroyRenderEnv2(li, tagid, false);
  return;
}

/*
 * DestroyRenderPart
 */
void
DestroyRenderPart(rp)
RenderPart *rp;
{
  if (rp->destroy != NULL) (*(rp->destroy))(rp->state);
  return;
}

/*
 * CreateRenderPart
 */
RenderPart *
CreateRenderPart(li, p, setup, render, destroy, select, state_size)
LayoutInfo *li;
MLTag *p;
void (*setup)();
void (*render)();
void (*destroy)();
void (*select)();
size_t state_size;
{
  RenderPart *rp;

  rp = (RenderPart *)MPCGet(li->mp, sizeof(RenderPart));
  rp->p = p;
  rp->anchorPart = li->restack->anchorPart;
  rp->namePart = li->restack->namePart;

  if (rp->anchorPart != NULL)
  {
    if (li->aphead == NULL) li->aphead = rp;
    else li->aptail->anext = rp;
    li->aptail = rp;
  }

  rp->setup = setup;
  rp->render = render;
  rp->destroy = destroy;
  rp->select = select;
  if (state_size > 0)
  {
    rp->state = (RenderPart *)MPCGet(li->mp, state_size);
  }
  rp->breakable = true;

  /*
   * Add to the master list.
   */
  if (li->rphead == NULL) li->rphead = rp;
  else li->rptail->mnext = rp;
  li->rptail = rp;

  return(rp);
}

/*
 * ComputeLine
 */
static void
ComputeLine(rl)
RenderLine *rl;
{
  RenderPart *rp;

  rl->width = 0;
  for (rp = rl->rphead; rp != NULL; rp = rp->lnext)
  {
    rp->x = rl->x + rl->width;
    rl->width += rp->width;
    if (rp->height > rl->maxy) rl->maxy = rp->height;
  }

  return;
}

/*
 * DisplayLine
 */
static void
DisplayLine(li, rl)
LayoutInfo *li;
RenderLine *rl;
{
  Region r;
  int x, y;
  unsigned int width, height;
  XRectangle rec;
  RenderPart *rp;

  r = XCreateRegion();
  FrameGetOnScreenDimensions(li->frame, &x, &y, &width, &height);
  rec.x = (short)x;
  rec.y = (short)y;
  rec.width = (unsigned short)width;
  rec.height = (unsigned short)height;
  XUnionRectWithRegion(&rec, r, r);

  for (rp = rl->rphead; rp != NULL; rp = rp->lnext)
  {
    if (IsRenderable(rp) && rp->render != NULL &&
        XRectInRegion(r, rp->x, rp->y, rp->width, rp->height))
    {
      (rp->render)(li, rp);
    }
  }

  XDestroyRegion(r);

  return;
}

/*
 * AddRenderPart
 */
int
AddRenderPart(li, rp, width, height)
LayoutInfo *li;
RenderPart *rp;
unsigned int width, height;
{
  RenderLine *rl = li->rltail;

  rp->width = width;
  rp->height = height;

  if (rl->rphead != NULL &&
      rl->width + rp->width >= rl->maxwidth &&
      li->restack->wrap)
  {
    RenderPart *xp, *fxp;
    RenderLine *orl;

    for (xp = rl->rptail; xp != NULL && !IsBreakable(xp); xp = xp->lprev)
	;

    if (xp != NULL)
    {
      orl = rl;
      rl = CreateRenderLine(li);

      if (xp->lnext != NULL)
      {
	fxp = xp->lnext;
	xp->lnext = NULL;
	
	ComputeLine(orl);
	
	for (; fxp->space; fxp = fxp->lnext)
	    ;
	
	fxp->lprev = NULL;
	rl->rphead = fxp;
	
	for (; fxp != NULL; fxp = fxp->lnext)
	{
	  rl->rptail = fxp;
	}
	
	ComputeLine(rl);
      }

      /* Paint the old line after everything is done shifting around */
      DisplayLine(li, orl);
    }
  }
  if (rl->rphead == NULL && rp->space) return(0);

  rp->x = rl->x + rl->width;
  rl->width += rp->width;
  if (rp->height > rl->maxy) rl->maxy = rp->height;

  if (rl->rptail != NULL) rl->rptail->lnext = rp;
  else rl->rphead = rp;
  rp->lprev = rl->rptail;
  rp->lnext = NULL;
  rl->rptail = rp;

  if (li->delay)
  {
    li->delay = false;
    LayoutHandler(li, NULL);
  }

  return(0);
}

/*
 * CreateRenderLine
 */
static RenderLine *
CreateRenderLine(li)
LayoutInfo *li;
{
  RenderLine *rl;
  RenderPart *rp;
  int adjustx;

  rl = (RenderLine *)MPCGet(li->mp, sizeof(RenderLine));
  rl->x = li->restack->x;
  rl->maxwidth = li->restack->width;
  rl->align = li->restack->align;

  if (li->rltail != NULL)
  {
    RenderLine *ll = li->rltail;

    ll->next = rl;
    if (ll->width > li->width)
    {
      li->width = ll->width;
      adjustx = 0;
    }

    if (ll->align == ALIGN_CENTER) adjustx = (ll->maxwidth - ll->width) / 2;
    else if (ll->align == ALIGN_RIGHT) adjustx = ll->maxwidth - ll->width;
    else adjustx = 0;

    ll->height = ll->maxy - ll->miny;
    if (ll->miny < 0) ll->y -= ll->miny;

    rl->y = li->line_space + ll->y + ll->height;
    li->height = ll->y + ll->height;

    for (rp = ll->rphead; rp != NULL; rp = rp->lnext)
    {
      rp->x += adjustx;
      if (rp->setup != NULL) (*(rp->setup))(li, rp, ll->y, ll->height);
    }
  }
  else
  {
    li->rlhead = rl;
    rl->y = li->margin;
  }

  li->rltail = rl;

  FrameSetSize(li->frame, true, li->width, li->height);

  return(rl);
}

/*
 * CallTagHandler
 */
static void
CallTagHandler(li, tag, p)
LayoutInfo *li;
int tag;
MLTag *p;
{
  int end;

  if (p == NULL) end = 1;
  else end = p->end;

  if (tag == TAG_EOF) HandleEOF(li);
  else if (tag == TAG_H1) HandleHx(li, tag, p, end);
  else if (tag == TAG_H2) HandleHx(li, tag, p, end);
  else if (tag == TAG_H3) HandleHx(li, tag, p, end);
  else if (tag == TAG_H4) HandleHx(li, tag, p, end);
  else if (tag == TAG_H5) HandleHx(li, tag, p, end);
  else if (tag == TAG_H6) HandleHx(li, tag, p, end);
  else if (tag == TAG_IMG) HandleImg(li, p, end);
  else if (tag == TAG_TITLE) HandleTitle(li, end);
  else if (tag == TAG_HR) HandleHR(li, p, end);
  else if (tag == TAG_P) HandleP(li, p, end);
  else if (tag == TAG_UL) HandleList(li, tag, p, end);
  else if (tag == TAG_OL) HandleList(li, tag, p, end);
  else if (tag == TAG_MENU) HandleList(li, tag, p, end);
  else if (tag == TAG_DIR) HandleList(li, tag, p, end);
  else if (tag == TAG_DL) HandleList(li, tag, p, end);
  else if (tag == TAG_LI) HandleLI(li, p, end);
  else if (tag == TAG_DT) HandleDT(li, p, end);
  else if (tag == TAG_DD) HandleDD(li, p, end);
  else if (tag == TAG_OL) HandleOL(li, p, end);
  else if (tag == TAG_BR) HandleBR(li, p, end);
  else if (tag == TAG_ADDRESS) HandleAddress(li, p, end);
  else if (tag == TAG_A) HandleA(li, p, end);
  else if (tag == TAG_PRE) HandlePRE(li, p, end);
  else if (tag == TAG_LISTING) HandleListing(li, p, end);
  else if (tag == TAG_PLAINTEXT) HandlePlaintext(li, p, end);
  else if (tag == TAG_B) HandleFontChange(li, end, tag);
  else if (tag == TAG_EM) HandleFontChange(li, end, tag);
  else if (tag == TAG_STRONG) HandleFontChange(li, end, tag);
  else if (tag == TAG_TT) HandleFontChange(li, end, tag);
  else if (tag == TAG_CITE) HandleFontChange(li, end, tag);
  else if (tag == TAG_I) HandleFontChange(li, end, tag);
  else if (tag == TAG_KBD) HandleFontChange(li, end, tag);
  else if (tag == TAG_FORM) HandleForm(li, p, end);
  else if (tag == TAG_INPUT) HandleInput(li, p, end);
  else if (tag == TAG_TEXTAREA) HandleTextarea(li, p, end);
  else if (tag == TAG_OPTION) HandleOption(li, p, end);
  else if (tag == TAG_SELECT) HandleSelect(li, p, end);
  else if (tag == TAG_BASE) HandleBase(li, p, end);
  else if (tag == TAG_HEAD) HandleMisc(li, tag, p, end);
  else if (tag == TAG_HTML) HandleMisc(li, tag, p, end);
  else if (tag == TAG_BODY) HandleMisc(li, tag, p, end);
  else if (tag == TAG_BLOCKQUOTE) HandleBlockquote(li, p, end);

  return;
}

/*
 * LayoutHandler
 */
void
LayoutHandler(li, p)
LayoutInfo *li;
MLTag *p;
{
  if (p != NULL)
  {
    if (li->pqhead == NULL) li->pqhead = p;
    else li->pqtail->qnext = p;
    li->pqtail = p;
  }

  for (; !li->delay && (p = li->pqhead) != NULL; )
  {
    li->pqhead = li->pqhead->qnext;

    if (li->restack->tagid == TAG_PLAINTEXT)
    {
      PlainText(li, p, true);
    }
    else if (p->id == TAG_TEXT)
    {
      switch (li->restack->text_mode)
      {
	case TAG_TEXTAREA:
          TextareaText(li, p);
	  break;
	case TAG_OPTION:
          OptionText(li, p);
	  break;
	case TAG_AU:
	  AuthorText(li, p);
	  break;
	case TAG_TITLE:
          TitleText(li, p);
	  break;
        case TAG_PRE:
	case TAG_LISTING:
	case TAG_CODE:
	  PlainText(li, p, false);
	  break;
        default:
	  FillText(li, p);
      }
    }
    else CallTagHandler(li, p->id, p);
  }

  return;
}
