/*
 * layout.h
 *
 * Copyright (c) 1995, John D. Kilburg <john@cs.unlv.edu>
 * See the file COPYRIGHT for details.
 */

#ifndef __LAYOUT_H__
#define __LAYOUT_H__ 1

#include <X11/Intrinsic.h>

#include "common.h"
#include "WWW.h"
#include "html.h"

#define FONT_SCALE_COUNT 6

/* Silly font information */
typedef struct _layout_font_info
{
  bool        fixed;            /* fixed width? */
  int         weight;           /* bold, medium, ... */
  int         slant;            /* italic, regular, ... */
  int         scale;            /* scale of font */
  int         size;             /* size of font */
  XFontStruct *xfi;             /* real font info */
} LayoutFontInfo;

/*
 * Helpful comment here
 */
typedef struct _render_env
{
  int tagid;                     /* tag ID that produced env */
  int x;                         /* x offsets */
  unsigned int width;            /* environment dimensions */
  int align;                     /* object alignment */
  int text_mode;                 /* fill, pre, title, ... */
  bool wrap;                     /* wrap lines automatically? */
  MLTag *anchorPart;             /* URL anchor */
  MLTag *namePart;               /* name anchor */
  LayoutFontInfo fi;             /* current font information */
  struct _render_env *next;      /* environment links */
} RenderEnv;

#define CurrentAlign(x)         ((x)->restack->align)

typedef struct _render_part
{
  MLTag *p;                      /* */
  MLTag *anchorPart;             /* URL anchor */
  MLTag *namePart;               /* name anchor */
  int x;                         /* set by AddRenderPart, CreateRenderLine */
  int y;                         /* set by render setup proc */
  unsigned int width, height;    /* dimensions including anchor stuff */
  bool breakable;                /* allow linebreak after this part ? */
  bool renderable;               /* can the part be rendered ? */
  bool space;                    /* space */
  struct _render_part *lnext, *lprev; /* line links */
  struct _render_part *mnext;    /* next in RP master list */
  struct _render_part *anext;    /* next in anchor list */

  /* handling functions */
  void (*setup)();               /* setup function */
  void (*render)();              /* render function */
  void (*destroy)();             /* destroy function */
  void (*select)();              /* select handler */
  void *state;                   /* state for functions */
} RenderPart;

#define GetRenderState(rp)       ((rp)->state)
#define IsRenderable(rp)         ((rp)->renderable)
#define IsBreakable(rp)          ((rp)->breakable)
#define IsSpace(rp)              ((rp)->space)

typedef struct _render_line
{
  RenderPart *rphead, *rptail;   /* list of RPs */
  int x, y;                      /* position */
  int miny, maxy;                /* offset from baseline? */
  unsigned int width, height;    /* height of line */
  unsigned int maxwidth;         /* maximum width of line */
  int align;                     /* alignment */
  struct _render_line *next;     /* next line */
} RenderLine;

typedef struct _layout_class
{
  /* font stuff */
  bool        font_setup_done;
  XFontStruct *fixedFont;
  XFontStruct *defaultFont;
  char        **fontNames;
  LayoutFontInfo *fontInfo;
  int         fontCount;
  int         scale[FONT_SCALE_COUNT];
} LayoutClass;

typedef struct _layout_info
{
  /* A whole wad of crazy stuff */
  Frame       frame;
  Widget      w;
  Display     *dpy;
  Window      win;
  GC          gc;
  Pixel       anchor_color;
  Pixel       fg, bg;
  MemPool     mp;
  MLState     *hs;
  Document    *d;
  LayoutClass *lc;
  LayoutFontInfo cfi;

  /* layout.c and other layout files */
  void        *formstate;           /* current form state */
  RenderPart  *aphead, *aptail;     /* RP anchor list */
  RenderPart  *rphead, *rptail;     /* RP master list */
  RenderLine  *rlhead, *rltail;     /* line list */
  RenderEnv   *restack;             /* environment stack */
  MLTag       *pqhead, *pqtail;     /* queue of unrendered tags */
  bool        delay;                /* delay layout ? */
  URLParts    *bup;                 /* base URL */
  unsigned int width, height;       /* dimensions of document */
  unsigned int margin;
  int          line_space;
  char        *title, *author;      /* title and author text */
  RenderPart  *over;

  /* other crap */
  bool        destroy_locked;       /* deals with destroy recursion problems */
  bool        reload;               /* reloading */
} LayoutInfo;

/*
 * Private functions
 */

/*
 * form.c
 */
void HandleForm _ArgProto((LayoutInfo *, MLTag *, int));
void HandleInput _ArgProto((LayoutInfo *, MLTag *, int));
void HandleTextarea _ArgProto((LayoutInfo *, MLTag *, int));
void TextareaText _ArgProto((LayoutInfo *, MLTag *));
void OptionText _ArgProto((LayoutInfo *, MLTag *));
void HandleSelect _ArgProto((LayoutInfo *, MLTag *, int));
void HandleOption _ArgProto((LayoutInfo *, MLTag *, int));

/*
 * inline.c
 */
void HandleImg _ArgProto((LayoutInfo *, MLTag *, int));

/*
 * hr.c
 */
void HandleHR _ArgProto((LayoutInfo *, MLTag *, int));

/*
 * list.c
 */
void HandleLI _ArgProto((LayoutInfo *, MLTag *, int));
void HandleDT _ArgProto((LayoutInfo *, MLTag *, int));
void HandleDD _ArgProto((LayoutInfo *, MLTag *, int));
void HandleOL _ArgProto((LayoutInfo *, MLTag *, int));
void HandleList _ArgProto((LayoutInfo *, int, MLTag *, int));

/*
 * text.c
 */
void AuthorText _ArgProto((LayoutInfo *, MLTag *));
void TitleText _ArgProto((LayoutInfo *, MLTag *));
void FillText _ArgProto((LayoutInfo *, MLTag *));
void PlainText _ArgProto((LayoutInfo *, MLTag *, bool));
void HandlePRE _ArgProto((LayoutInfo *, MLTag *, int));
void HandleListing _ArgProto((LayoutInfo *, MLTag *, int));
void HandlePlaintext _ArgProto((LayoutInfo *, MLTag *, int));
void HandleTitle _ArgProto((LayoutInfo *, int));

/*
 * layout.c
 */
void LayoutHandler _ArgProto((LayoutInfo *, MLTag *));
void AddBlankLine _ArgProto((LayoutInfo *));
bool InEnv _ArgProto((LayoutInfo *, int));
void LayoutSpacify _ArgProto((char *, size_t));
URLParts *GetBaseURLParts _ArgProto((LayoutInfo *));
int GetWidthHeight _ArgProto((MLTag *, unsigned int *, unsigned int *));

RenderPart *CreateRenderPart _ArgProto((LayoutInfo *, MLTag *,
					void (*)(),
					void (*)(),
					void (*)(),
					void (*)(),
					size_t));
int AddRenderPart _ArgProto((LayoutInfo *, RenderPart *,
			     unsigned int, unsigned int));
void DestroyRenderPart _ArgProto((RenderPart *));

RenderEnv *CreateRenderEnv _ArgProto((LayoutInfo *, int));
void DestroyRenderEnv _ArgProto((LayoutInfo *, int));
void DestroyRenderEnv2 _ArgProto((LayoutInfo *, int, bool));
void AddLineBreak _ArgProto((LayoutInfo *));
unsigned int RemainingLineWidth _ArgProto((LayoutInfo *));

/*
 * misc.c
 */
void HandleHx _ArgProto((LayoutInfo *, int, MLTag *, int));
void HandleP _ArgProto((LayoutInfo *, MLTag *, int));
void HandleBR _ArgProto((LayoutInfo *, MLTag *, int));
void HandleAddress _ArgProto((LayoutInfo *, MLTag *, int));
void HandleA _ArgProto((LayoutInfo *, MLTag *, int));
void HandleBase _ArgProto((LayoutInfo *, MLTag *, int));
void HandleBlockquote _ArgProto((LayoutInfo *, MLTag *, int));
void HandleFontChange _ArgProto((LayoutInfo *, int, int));
void HandleMisc _ArgProto((LayoutInfo *, int, MLTag *, int));
void HandleEOF _ArgProto((LayoutInfo *));

/*
 * font.c
 */
XFontStruct *LayoutGetFont _ArgProto((LayoutInfo *));
void LayoutSetupFonts _ArgProto((LayoutInfo *));

#endif
