/*
 * text.c
 *
 * Copyright (c) 1995, John D. Kilburg (john@cs.unlv.edu)
 * See the file COPYRIGHT for details.
 */
#include <stdio.h>
#include <ctype.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include "layout.h"

#define TAB_EXPANSION 4

typedef struct _text_state
{
  char *s;
  size_t len;
  XFontStruct *font;
} TextState;

/*
 *
 * Private functions
 *
 */
static void RenderText _ArgProto((LayoutInfo *, RenderPart *));
static RenderPart *CreateTextState _ArgProto((LayoutInfo *,
					      MLTag *p,
					      char *, size_t,
					      XFontStruct *));
static void SetupText _ArgProto((LayoutInfo *, RenderPart *, int, int));
static char *ExpandTabs _ArgProto((LayoutInfo *, char *, size_t, size_t));

/*
 * SetupText
 */
static void
SetupText(li, rp, line_y, line_height)
LayoutInfo *li;
RenderPart *rp;
int line_y;
int line_height;
{
  rp->y = line_y + line_height - rp->height;
  rp->renderable = true;
  return;
}

/*
 * RenderText
 */
static void
RenderText(li, rp)
LayoutInfo *li;
RenderPart *rp;
{
  TextState *ts = (TextState *)GetRenderState(rp);

  XSetFont(li->dpy, li->gc, ts->font->fid);

  if (rp->anchorPart != NULL)
  {
    XSetForeground(li->dpy, li->gc, li->anchor_color);
    XDrawString(li->dpy, li->win, li->gc,
		rp->x, rp->y + ts->font->ascent, ts->s, (int)ts->len);
    XDrawLine(li->dpy, li->win, li->gc,
	      rp->x, rp->y + rp->height,
	      rp->x + rp->width, rp->y + rp->height);
    XSetForeground(li->dpy, li->gc, li->fg);
  }
  else
  {
    XDrawString(li->dpy, li->win, li->gc,
		rp->x, rp->y + ts->font->ascent, ts->s, (int)ts->len);
  }

  return;
}

/*
 * CreateTextState
 */
static RenderPart *
CreateTextState(li, p, s, slen, font)
LayoutInfo *li;
MLTag *p;
char *s;
size_t slen;
XFontStruct *font;
{
  RenderPart *rp;
  TextState *ts;

  rp = CreateRenderPart(li, p, SetupText,
			RenderText, NULL, NULL, sizeof(TextState));
  ts = (TextState *)GetRenderState(rp);
  ts->s = s;
  ts->len = slen;
  ts->font = font;

  if (s == NULL)
  {
    s = " ";
    slen = 1;
    rp->space = true;
  }
  else rp->breakable = false;

  AddRenderPart(li, rp,
		XTextWidth(ts->font, s, slen), font->ascent + font->descent);

  return(rp);
}

/*
 *
 * Public functions
 *
 */

/*
 * TitleText
 */
void
TitleText(li, p)
LayoutInfo *li;
MLTag *p;
{
  size_t title_len;
  char *text;
  size_t len;

  MLGetTagText(li->hs, p, false, &text, &len);
  if (text == NULL) return;

  if (li->title == NULL)
  {
    li->title = (char *)alloc_mem(len + 1);
    title_len = 0;
  }
  else
  {
    title_len = strlen(li->title);
    li->title = (char *)realloc_mem(li->title, title_len + len + 1);
  }

  strncpy(li->title + title_len, text, len);
  li->title[title_len + len] = '\0';

  LayoutSpacify(li->title, title_len + len);

  return;
}

/*
 * AuthorText
 */
void
AuthorText(li, p)
LayoutInfo *li;
MLTag *p;
{
  size_t author_len;
  char *text;
  size_t len;

  MLGetTagText(li->hs, p, false, &text, &len);
  if (text == NULL) return;

  if (li->author == NULL)
  {
    li->author = (char *)alloc_mem(len + 1);
    author_len = 0;
  }
  else
  {
    author_len = strlen(li->author);
    li->author = (char *)realloc_mem(li->author, author_len + len + 1);
  }

  strncpy(li->author + author_len, text, len);
  li->author[author_len + len] = '\0';

  return;
}

/*
 * ExpandTabs
 */
static char *
ExpandTabs(li, t, len, tlen)
LayoutInfo *li;
char *t;
size_t len;
size_t tlen;
{
  char *s, *cp, *lastcp, *xs;

  lastcp = t + len;
  s = (char *)MemPoolGet(li->mp, tlen);
  for (cp = t, xs = s; cp < lastcp; cp++)
  {
    if (*cp == '\t')
    {
      memset(xs, ' ', TAB_EXPANSION);
      xs += TAB_EXPANSION;
    }
    else if (*cp != '\r')
    {
      *xs = *cp;
      xs++;
    }
  }

  return(s);
}

/*
 * PlainText
 */
void
PlainText(li, p, original)
LayoutInfo *li;
MLTag *p;
bool original;
{
  char *cp, *lastcp, *start;
  size_t ptlen;
  XFontStruct *font;
  char *s;
  char *text;
  size_t len;
  bool tabbed;

  MLGetTagText(li->hs, p, original, &text, &len);
  if (text == NULL) return;

  font = LayoutGetFont(li);

  lastcp = text + len;
  tabbed = false;
  for (ptlen = 0, start = cp = text; cp < lastcp; cp++)
  {
    if (*cp == '\r') continue;
    else if (*cp == '\n')
    {
      if (tabbed)
      {
	s = ExpandTabs(li, start, cp - start, ptlen);
	CreateTextState(li, p, s, ptlen, font);
      }
      else
      {
	CreateTextState(li, p, start, ptlen, font);
      }
      AddLineBreak(li);

      ptlen = 0;
      start = cp + 1;
      tabbed = false;
    }
    else if (*cp == '\t')
    {
      ptlen += TAB_EXPANSION;
      tabbed = true;
    }
    else ptlen++;
  }
  
  if (tabbed)
  {
    s = ExpandTabs(li, start, cp - start, ptlen);
    CreateTextState(li, p, s, ptlen, font);
  }
  else
  {
    CreateTextState(li, p, start, ptlen, font);
  }
  
  return;
}

/*
 * FillText
 */
void
FillText(li, p)
LayoutInfo *li;
MLTag *p;
{
  char *s, *cp, *lastcp;
  XFontStruct *font;
  bool found_space;
  RenderPart *rp;
  char *text;
  size_t len;

  MLGetTagText(li->hs, p, false, &text, &len);
  if (text == NULL) return;

  font = LayoutGetFont(li);
  for (lastcp = text + len, cp = text; cp < lastcp; )
  {
    for (found_space = false; cp < lastcp && isspace(*cp); cp++)
    {
      found_space = true;
    }
    
    if (found_space) rp = CreateTextState(li, p, NULL, 0, font);
    if (cp == lastcp) break;

    for (s = cp; cp < lastcp && !isspace(*cp); cp++)
	;

    rp = CreateTextState(li, p, s, cp - s, font);
  }

  return;
}

void
HandleTitle(li, end)
LayoutInfo *li;
int end;
{
  if (end) DestroyRenderEnv(li, TAG_TITLE);
  else
  {
    RenderEnv *re;

    re = CreateRenderEnv(li, TAG_TITLE);
    re->text_mode = TAG_TITLE;
  }

  return;
}

void
HandlePRE(li, p, end)
LayoutInfo *li;
MLTag *p;
int end;
{
  if (end)
  {
    AddBlankLine(li);
    DestroyRenderEnv(li, TAG_PRE);
    AddLineBreak(li);
  }
  else
  {
    RenderEnv *re;

    AddLineBreak(li);
    re = CreateRenderEnv(li, TAG_PRE);
    re->fi.fixed = true;
    re->text_mode = TAG_PRE;
    re->wrap = false;
    AddBlankLine(li);
  }

  return;
}

void
HandleListing(li, p, end)
LayoutInfo *li;
MLTag *p;
int end;
{
  if (end)
  {
    AddBlankLine(li);
    DestroyRenderEnv(li, TAG_LISTING);
    AddLineBreak(li);
  }
  else
  {
    RenderEnv *re;

    AddLineBreak(li);
    re = CreateRenderEnv(li, TAG_LISTING);
    re->fi.fixed = true;
    re->text_mode = TAG_LISTING;
    re->wrap = false;
    AddBlankLine(li);
  }

  return;
}

void
HandlePlaintext(li, p, end)
LayoutInfo *li;
MLTag *p;
int end;
{
  if (end)
  {
    AddBlankLine(li);
    DestroyRenderEnv(li, TAG_PLAINTEXT);
    AddLineBreak(li);
  }
  else
  {
    RenderEnv *re;

    AddLineBreak(li);
    re = CreateRenderEnv(li, TAG_PLAINTEXT);
    re->fi.fixed = true;
    re->wrap = false;
    AddBlankLine(li);
  }

  return;
}



