/*
 * jpeg.c:
 *
 * (c) 1995 Erik Corry ehcorry@inet.uni-c.dk
 * See the file COPYRIGHT for details
 *
 * modelled on xbm.c.
 *
 * If you want to understand this read the following files from libjpeg:
 *
 *   libjpeg.doc
 *   jdatasrc.c
 *   jpeglib.h
 *   example.c
 *
 */

#include <stdlib.h>
#include <string.h>

#include "common.h"
#include "imagep.h"
#include "jpegp.h"

/*
 * jpegGetImage - called through vector
 */

static Image *
jpegGetImage(void *pointer)
{
  jpegState *jpeg = (jpegState *)pointer;
  return jpeg->image;
}

/*
 * jpegDestroy - called through vector
 */
static void
jpegDestroy(pointer)
void *pointer;
{
  jpegState *jpeg = (jpegState *)pointer;
  if(setjmp(jpeg->error_state.jump_buffer))
  {
    /*
     * Oh shit, an error while clearing up...
     */
    return;
  }
  if (jpeg->image)
    freeImage(jpeg->image);
  jpeg->image = 0;
  if (jpeg->destroy_jpeg)
    jpeg_destroy_decompress(&jpeg->cinfo);
  jpeg->destroy_jpeg = false;
  if (jpeg != NULL) free_mem(jpeg);

  return;
}

static void
lf_init_source_callback(
  struct jpeg_decompress_struct *cinfo)
{
}

static boolean
lf_fill_input_buffer_callback(
  struct jpeg_decompress_struct *cinfo)
{
  struct jpeg_chimera_input_state *input_state =
    (struct jpeg_chimera_input_state *)(cinfo->src);
  

  if(input_state->old_len >= input_state->len)
    return FALSE;                 /* There is no more data - suspend IO */

  input_state->pub.next_input_byte = input_state->data + input_state->old_len;
  input_state->pub.bytes_in_buffer = input_state->len - input_state->old_len;
  input_state->old_len = input_state->len;

  return TRUE;
}

static void
lf_skip_input_data_callback(
  struct jpeg_decompress_struct *cinfo,
  long num_bytes)
{
  struct jpeg_chimera_input_state *input_state =
    (struct jpeg_chimera_input_state *)(cinfo->src);
  

  if(num_bytes <= 0) return;

  if(num_bytes > input_state->pub.bytes_in_buffer)
  {
    input_state->old_len += num_bytes - input_state->pub.bytes_in_buffer;
    input_state->pub.bytes_in_buffer = 0;
  }
  else
  {
    input_state->pub.next_input_byte += num_bytes;
    input_state->pub.bytes_in_buffer -= num_bytes;
  }
}

static void
lf_term_source_callback(
  struct jpeg_decompress_struct *cinfo)
{
}

static void
lf_error_exit_callback(
  j_common_ptr cinfo)
{
  struct jpeg_chimera_error_state *error_state =
    (struct jpeg_chimera_error_state *)(cinfo->err);
  /*
   * Display message. At the moment I use the builtin stderr handler,
   * but we could install something smarter, or just keep quiet about
   * it.
   */
  (error_state->pub.output_message)(cinfo);

  /* Return control to the setjmp point */
  longjmp(error_state->jump_buffer, 1);
}

static int
lf_read_header(jpegState *jpeg)
{
  int status = jpeg_read_header(&jpeg->cinfo, TRUE);

  if(status == JPEG_SUSPENDED) return CH_JPEG_NEED_DATA;

  /*
   * If this worked, set decompression parameters
   */

  if(jpeg->cinfo.progressive_mode)
    jpeg->cinfo.buffered_image = TRUE;
  
  jpeg->state = CH_JPEG_START_DECOMPRESS;
  return CH_JPEG_SUCCESS;
}

static int
lf_start_decompress(jpegState *jpeg)
{
  int status = jpeg_start_decompress(&jpeg->cinfo);

  if(status == FALSE) return CH_JPEG_NEED_DATA;

  /*
   * If this worked, get image parameters
   */
 if(jpeg->cinfo.output_components == 1)
 {
   int i;
   jpeg->image = newRGBImage(jpeg->cinfo.output_width,
                             jpeg->cinfo.output_height, 8);
   jpeg->image->type = IGRAY;
   jpeg->image->rgb.red = jpeg->cmap;
   jpeg->image->rgb.green = jpeg->cmap;
   jpeg->image->rgb.blue = jpeg->cmap;
   jpeg->image->rgb.used = 256;
   for(i = 0; i < 256; i++)
     jpeg->cmap[i] = i | (i << 8);
 }
 else
 {
   jpeg->image = newTrueImage(jpeg->cinfo.output_width,
                              jpeg->cinfo.output_height);
 } 

  /*
   * Next state
   */
  jpeg->state = CH_JPEG_START_OUTPUT;
  return CH_JPEG_SUCCESS;
}

static int
lf_start_output(jpegState *jpeg)
{
  if(jpeg->cinfo.buffered_image)
  {
    int status = JPEG_REACHED_SOS;
    while(status != JPEG_SUSPENDED && status != JPEG_REACHED_EOI)
      status = jpeg_consume_input(&jpeg->cinfo);
    jpeg_start_output(&jpeg->cinfo, jpeg->cinfo.input_scan_number);
    /* printf("scan %d\n", jpeg->cinfo.input_scan_number); */
  }

  jpeg->ypos = 0;
  jpeg->state = CH_JPEG_READ_IMAGE;
  return CH_JPEG_SUCCESS;
}

static void
lf_destroy_decompress(jpegState *jpeg)
{
  jpeg_destroy_decompress(&jpeg->cinfo);
  jpeg->destroy_jpeg = false;
}

static int
lf_read_image(jpegState *jpeg)
{
  int maxlines;
  int scanline_count;
  unsigned char *scanline_pointers[32];
  unsigned char *output_addr;

  while(jpeg->ypos < jpeg->image->height)
  {
    output_addr =
      jpeg->image->data + jpeg->image->bytes_per_line * jpeg->ypos;
  
    for(maxlines = 0;
        maxlines < 32 && maxlines + jpeg->ypos < jpeg->image->height;
        maxlines++)
    {
      scanline_pointers[maxlines] = output_addr;
      output_addr += jpeg->image->bytes_per_line;
    }
  
    scanline_count =
      jpeg_read_scanlines(&jpeg->cinfo, scanline_pointers, maxlines);
  
    if(scanline_count == 0) return CH_JPEG_NEED_DATA;
  
    if (jpeg->lineProc != NULL)
      (jpeg->lineProc)(jpeg->closure, jpeg->ypos,
                       jpeg->ypos + scanline_count - 1);
  
    jpeg->ypos += scanline_count;
  }

  jpeg->state = CH_JPEG_FINISH_DECOMPRESS;
  return CH_JPEG_SUCCESS;
}

static int
lf_finish_decompress(jpegState *jpeg)
{
  int status;

  if(jpeg->cinfo.buffered_image)
  {
    status = jpeg_finish_output(&jpeg->cinfo);
    if (status == FALSE) return CH_JPEG_NEED_DATA;

    if(!jpeg_input_complete(&jpeg->cinfo))
    {
      jpeg->state = CH_JPEG_START_OUTPUT;
      return CH_JPEG_SUCCESS;
    }
  }

  status = jpeg_finish_decompress(&jpeg->cinfo);
  if (status == FALSE) return CH_JPEG_NEED_DATA;
  lf_destroy_decompress(jpeg);
  jpeg->state = CH_JPEG_FINISHED;
  return CH_JPEG_SUCCESS;
}

/*
 * jpegAddData - called through vector
 *
 * 0 success
 * 1 needs more data
 * -1 error
 *
 * Assumes data is the address of the beginning of the jpeg data and len
 * is the total length.
 */
static int
jpegAddData(pointer, data, len, data_ended)
void *pointer;
byte *data;
int len;
bool data_ended;
{
  jpegState *jpeg = (jpegState *)pointer;
  int rval;

  jpeg->input_state.data = data;
  jpeg->input_state.len = len;

  if(setjmp(jpeg->error_state.jump_buffer))
  {
    /* jpegDestroy((void *)jpeg); */
    return -1;
  }

  if(len <= jpeg->input_state.old_len)
  {
    if (data_ended) return(-1);
    return 1;
  }

  /*
   * We want at least 2048 bytes at a time unless the file has ended
   */
  if(len - jpeg->input_state.old_len < 2048 && !data_ended)
    return 1;

  for ( ; ; )
  {
    switch (jpeg->state)
    {
      case CH_JPEG_READ_HEADER:
        rval = lf_read_header(jpeg);
        break;
      case CH_JPEG_START_DECOMPRESS:
        rval = lf_start_decompress(jpeg);
	break;
      case CH_JPEG_START_OUTPUT:
        rval = lf_start_output(jpeg);
	break;
      case CH_JPEG_READ_IMAGE:
        rval = lf_read_image(jpeg);
	break;
      case CH_JPEG_FINISH_DECOMPRESS:
        rval = lf_finish_decompress(jpeg);
	break;
      case CH_JPEG_FINISHED:
        return 0;
    }
    if (rval == CH_JPEG_NEED_DATA)
    {
      if (data_ended) return(-1);
      return(1);
    }
    else if (rval != CH_JPEG_SUCCESS) return(-1);
  }

  return(-1);
}

/*
 * jpegInit
 *
 * Initialize JPEG reader state
 */
void
jpegInit(lineProc, closure, if_vector)
void (*lineProc)();
void *closure;
struct ifs_vector *if_vector;
{
  jpegState *jpeg;

  jpeg = (jpegState *)alloc_mem(sizeof(jpegState));
  memset(jpeg, 0, sizeof(jpegState));
  jpeg->state = CH_JPEG_READ_HEADER;
  jpeg->lineProc = lineProc;
  jpeg->closure = closure;

  /*
   * Initialise error handler.
   * We set up the normal JPEG error routines, then override error_exit.
   */
  jpeg->cinfo.err = jpeg_std_error(&jpeg->error_state.pub);
  jpeg->cinfo.err->error_exit = &lf_error_exit_callback;
  if(setjmp(jpeg->error_state.jump_buffer))
  {
    /*
     * We get here after a failure in initialisation. Lets just assume that
     * people are careful about the order in which things are done, and that
     * a destroy will work here.
     */
    jpegDestroy((void *)jpeg);
    /*
     * It would be nice if init could indicate failure...
     */
    return;
  }

  /*
   * Initialise jpeg library
   */
  jpeg_create_decompress(&jpeg->cinfo);
  jpeg->destroy_jpeg = true;

  /*
   * Initialise input. This is really messy. What we do is make a pointer
   * to our own structure, but at the start of our own structure there
   * is the structure the library expects, so its all right...
   */

  jpeg->cinfo.src = &jpeg->input_state.pub;
  jpeg->cinfo.src->init_source = &lf_init_source_callback;
  jpeg->cinfo.src->fill_input_buffer = &lf_fill_input_buffer_callback;
  jpeg->cinfo.src->skip_input_data = &lf_skip_input_data_callback;
  jpeg->cinfo.src->resync_to_restart = &jpeg_resync_to_restart;
  jpeg->cinfo.src->term_source = &lf_term_source_callback;
  jpeg->cinfo.src->bytes_in_buffer = 0;
  jpeg->cinfo.src->next_input_byte = NULL;

  jpeg->input_state.old_len = 0;

  /*
   * Initialise vector for methods Chimera needs
   */

  if_vector->initProc = &jpegInit;
  if_vector->destroyProc = &jpegDestroy;
  if_vector->addDataProc = &jpegAddData;
  if_vector->getImageProc = &jpegGetImage;

  if_vector->image_format_closure = (void *)jpeg;
}
