/*
 * cache.c
 *
 * Copyright (C) 1995, John Kilburg <john@cs.unlv.edu>
 * See the file COPYRIGHT for details.
 */
#include <stdio.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>

#if defined(SYSV) || defined(SVR4) || defined(__arm)
#include <dirent.h>
#define DIRSTUFF struct dirent
#else
#include <sys/dir.h>
#define DIRSTUFF struct direct
#endif

#include "WWW.h"

typedef struct _cache_entry
{
  MemPool mp;
  char *hash;                         /* URL hash */
  char *filename;
  FILE *fp;
  off_t wlen;
  bool written;
  struct _cache_class *cc;
  struct _cache_entry *next;
} CacheEntry;

typedef struct _cache_class
{
  MemPool mp;
  char *dirname;
  CacheEntry *clist;
  bool persist;
  size_t max_size;
  size_t curr_size;
} CacheClass;

typedef struct _cr_info
{
  MemPool mp;

  CacheClass *cc;
  CacheEntry *ce;
  FILE *fp;
  Frame frame;
  IOTimeOut iot;
  off_t dstart;
  off_t rlen;
  Document *d;
  bool init_done;

  /* message stuff */
  char *url;
  char *mbuf;
  int mlen;
  char *rstr;
} CRInfo;

typedef struct _cw_info
{
  MemPool mp;
  CacheClass *cc;
  CacheEntry *ce;
  Frame frame;
} CWInfo;

static CacheEntry *CacheEntryFind _ArgProto((char *, CacheClass *));
static void CacheEntryDestroy _ArgProto((CacheEntry *, bool));
static CacheEntry *CacheEntryCreate _ArgProto((CacheClass *, char *));
static char *MakeMessage _ArgProto((CRInfo *, char *, size_t));

/*
 * MakeMessage
 */
static char *
MakeMessage(cr, str, n)
CRInfo *cr;
char *str;
size_t n;
{
  size_t len;

  len = strlen(str) + strlen(cr->url) + 100 + 100;
  if (len > cr->mlen)
  {
    cr->mbuf = (char *)MPGet(cr->mp, len);
    cr->mlen = len;
  }
  sprintf (cr->mbuf, str, n, cr->url);

  return(cr->mbuf);
}

/*
 * CacheEntryFind
 */
static CacheEntry *
CacheEntryFind(hash, cc)
char *hash;
CacheClass *cc;
{
  CacheEntry *ce;

  for (ce = cc->clist; ce != NULL; ce = ce->next)
  {
    if (strcmp(hash, ce->hash) == 0)
    {
      if (ce->written) return(ce);
      else return(NULL);
    }
  }

  return(NULL);
}

/*
 * CacheEntryDestroy
 */
static void
CacheEntryDestroy(ce, persist)
CacheEntry *ce;
bool persist;
{
  if (!persist && ce->filename != NULL) unlink(ce->filename);
  if (ce->fp != NULL) fclose(ce->fp);
  MPDestroy(ce->mp);
  return;
}

/*
 * CacheEntryCreate
 */
static CacheEntry *
CacheEntryCreate(cc, hash)
CacheClass *cc;
char *hash;
{
  CacheEntry *ce = NULL;
  char *filename;
  FILE *fp;
  MemPool mp;

  mp = MPCreate();
  filename = (char *)MPGet(mp, strlen("ccf") +
			   strlen(cc->dirname) +
			   strlen(hash) + strlen("/") + 1);
  strcpy(filename, cc->dirname);
  strcat(filename, "/");
  strcat(filename, "ccf");
  strcat(filename, hash);

  if ((fp = fopen(filename, "w")) == NULL)
  {
    MPDestroy(mp);
    return(NULL);
  }

  ce = (CacheEntry *)MPCGet(mp, sizeof(CacheEntry));
  ce->mp = mp;
  ce->next = cc->clist;
  cc->clist = ce;
  ce->hash = MPStrDup(mp, hash);
  ce->filename = filename;
  ce->fp = fp;
  ce->cc = cc;

  return(ce);
}

/*
 * CRDestroy
 */
static int
CRDestroy(cr)
CRInfo *cr;
{
  if (cr->fp != NULL) fclose(cr->fp);
  if (cr->iot != NULL) IODestroyTimeOut(cr->iot);
  if (!cr->init_done && cr->d != NULL) DocumentDestroy(cr->d);
  MPDestroy(cr->mp);
  return(0);
}

/*
 * CRRead
 */
static void
CRRead(iot, cr)
IOTimeOut iot;
CRInfo *cr;
{
  off_t rval;
  Document *d = cr->d;

  rval = (off_t)fread(d->ds + cr->rlen, 1, BUFSIZ, cr->fp);
  if (rval == 0)
  {
    if (cr->init_done)
    {
      d->data = d->ds + cr->dstart;
      d->len = cr->rlen - cr->dstart;
      FrameEndData(cr->frame, d);
    }
  }
  else
  {
    cr->rlen += rval;
    
    if (d->mflist == NULL)
    {
      if ((cr->dstart = MIMEFindData(d->ds, cr->rlen)) != -1)
      {
	d->mflist = MIMEParseHeader(d->mp, d->ds, cr->dstart);
	cr->init_done = true;
	FrameInitData(cr->frame, d);
      }
    }
    
    if (d->mflist != NULL)
    {
      d->data = d->ds + cr->dstart;
      d->len = cr->rlen - cr->dstart;
      FrameAddData(cr->frame, d);
      IOSetTimeOut(cr->iot, 10, CRRead);
    }
  }

  FrameMessage(cr->frame, MakeMessage(cr, cr->rstr, d->len));

  return;
}

/*
 * CRInit
 */
static void *
CRInit(up, pup, cc, frame)
URLParts *up;
URLParts *pup;
CacheClass *cc;
Frame frame;
{
  CRInfo *cr;
  struct stat s;
  CacheEntry *ce;
  char *dbstr;
  MemPool mp;
  char *hash;

  mp = MPCreate();
  hash = URLHash(mp, up);
  if ((ce = CacheEntryFind(hash, cc)) == NULL)
  {
    MPDestroy(mp);
    return(NULL);
  }

  cr = (CRInfo *)MPCGet(mp, sizeof(CRInfo));
  cr->mp = mp;
  cr->frame = frame;
  cr->ce = ce;
  cr->url = URLMakeString(cr->mp, up, false);

  if (ce->filename == NULL || stat(ce->filename, &s) == -1 ||
      (cr->fp = fopen(ce->filename, "r")) == NULL)
  {
    CRDestroy(cr);
    return(NULL);
  }
  
  dbstr = FrameGetResource(frame, "cache.found");
  if (dbstr == NULL) dbstr = "cache.found";
  cr->rstr = FrameGetResource(frame, "cache.read");
  if (cr->rstr == NULL) cr->rstr = "cache.read";
  
  cr->mlen = strlen(dbstr) + strlen(cr->url) + BUFSIZ;
  cr->mbuf = (char *)MPGet(cr->mp, cr->mlen);
  sprintf (cr->mbuf, dbstr, cr->url);
  FrameMessage(frame, cr->mbuf);

  cr->d = DocumentCreate(false);
  cr->d->ds = (byte *)alloc_mem(s.st_size);
  cr->iot = IOCreateTimeOut(frame, cr);
  IOSetTimeOut(cr->iot, 10, CRRead);

  return(cr);
}

/*
 * Cache Writing
 */

/*
 * CWDestroy
 */
static int
CWDestroy(cwi)
CWInfo *cwi;
{
  MPDestroy(cwi->mp);
  return(0);
}

/*
 * CWAdd
 */
static int
CWAdd(cwi, d)
CWInfo *cwi;
Document *d;
{
  CacheEntry *ce = cwi->ce;

  if (ce->fp == NULL) return(0);
  if (ce->wlen < d->len)
  {
    int rval;

    rval = fwrite(d->data + ce->wlen, 1, d->len - ce->wlen, ce->fp);
    if (rval == 0)
    {
      fclose(ce->fp);
      ce->fp = NULL;
    }
    else ce->wlen += rval;
  }

  return(0);
}

/*
 * CWEnd
 */
static int
CWEnd(cwi, d)
CWInfo *cwi;
Document *d;
{
  CacheEntry *ce = cwi->ce;

  if (ce->fp == NULL) return(0);
  if (ce->wlen < d->len)
  {
    fwrite(d->data + ce->wlen, 1, d->len - ce->wlen, ce->fp);
  }
  fclose(ce->fp);
  ce->fp = NULL;

  ce->written = true;
  ce->cc->curr_size += d->len;

  return(0);
}

/*
 * CWInit
 */
static void *
CWInit(up, d, cc, frame)
URLParts *up;
Document *d;
CacheClass *cc;
Frame frame;
{
  MIMEField *mf;
  CWInfo *cwi;
  CacheEntry *ce;
  MemPool mp;
  char *hash;

  /*
   * Don't cache authenticated documents.  Don't cache documents that
   * require input.  Don't cache documents that have been marked 'nocache'
   */
  if (up->username != NULL || up->password != NULL ||
      up->input_data != NULL || d->nocache)
  {
    return(NULL);
  }

  mp = MPCreate();
  hash = URLHash(mp, up);

  /*
   * If a cache entry already exists, don't cache.
   */
  if (CacheEntryFind(hash, cc) != NULL)
  {
    MPDestroy(mp);
    return(NULL);
  }

  if ((ce = CacheEntryCreate(cc, hash)) == NULL)
  {
    MPDestroy(mp);
    return(NULL);
  }

  cwi = (CWInfo *)MPCGet(mp, sizeof(CWInfo));
  cwi->mp = mp;
  cwi->cc = cc;
  cwi->ce = ce;
  cwi->frame = frame;

  for (mf = d->mflist; mf != NULL; mf = mf->next)
  {
    fprintf (ce->fp, "%s: %s\n", mf->name, mf->value);
  }
  fprintf (ce->fp, "\n");

  return(cwi);
}

/*
 * CacheClassDestroy
 */
static int
CacheClassDestroy(cc)
CacheClass *cc;
{
  CacheEntry *c, *t;

  for (c = cc->clist; c != NULL; )
  {
    t = c;
    c = c->next;
    CacheEntryDestroy(t, cc->persist);
  }
  MPDestroy(cc->mp);

  return(0);
}

/*
 * InitModule_Cache
 */
void
InitModule_Cache(wc)
WWWContext wc;
{
  static ProtoHooks ph;
  static RenderHooks rh;
  CacheClass *cc;
  char *dirname;
  char *max;
  DIRSTUFF *de;
  DIR *dp;
  MemPool mp;
  
  if ((dirname = WWWGetResource(wc, "cache.directory")) == NULL) return;
  
  mp = MPCreate();
  cc = (CacheClass *)MPCGet(mp, sizeof(CacheClass));
  cc->mp = mp;
  
  cc->dirname = FixPath(mp, dirname);
  
  if ((dp = opendir(dirname)) != NULL)
  {
    while ((de = readdir(dp)) != NULL)
    {
      if (strncmp("ccf", de->d_name, 3) == 0)
      {
	CacheEntryCreate(cc, de->d_name + 3);
      }
    }
    
    closedir(dp);
  }
  
  if (WWWGetResource(wc, "cache.persist") != NULL) cc->persist = true;
  if ((max = WWWGetResource(wc, "cache.maxSize")) != NULL)
  {
    cc->max_size = atoi(max);
  }
  if (cc->max_size == 0) cc->max_size = 3000000;

  memset(&ph, 0, sizeof(ProtoHooks));
  ph.protocol = "cache-read";
  ph.init = CRInit;
  ph.destroy = CRDestroy;
  ph.class_context = cc;
  WWWAddProtoHooks(wc, &ph);
  
  memset(&rh, 0, sizeof(RenderHooks));
  rh.content = "x-www/x-cache-write";
  rh.init = CWInit;
  rh.destroy = CWDestroy;
  rh.add = CWAdd;
  rh.end = CWEnd;
  rh.class_context = cc;
  rh.class_destroy = CacheClassDestroy;
  WWWAddRenderHooks(wc, &rh);

  return;
}
