/*
 * download.c
 *
 * Copyright (C) 1995, John Kilburg <john@cs.unlv.edu>
 * See the file COPYRIGHT for details.
 */
#include <stdio.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Command.h>
#include "ScrollText.h"

#include "WWW.h"

typedef struct _download_info
{
  MemPool mp;
  Frame frame;
  URLParts *up;
  Widget w, tw;
  Widget save;
  Frame message_frame;
  FILE *fp;
  bool tried;
  bool finished;
  size_t cwlen;
  Document *d;
} DownloadInfo;

/*
 * DownloadSaveCallback
 */
static void
DownloadSaveCallback(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  DownloadInfo *di = (DownloadInfo *)cldata;
  char *filename;

  if (di->fp != NULL) return;
  XtVaGetValues(di->tw, XtNstring, &filename, NULL);
  if (filename == NULL || filename[0] == '\0') return;
  di->fp = fopen(filename, "w");

  if (di->finished)
  {
    fwrite(di->d->data, 1, di->d->len, di->fp);
    fclose(di->fp);
    di->fp = NULL;
  }

  return;
}

/*
 * DownloadAddData
 */
static int
DownloadAddData(di, d)
DownloadInfo *di;
Document *d;
{
  int rval;

  if (di->fp == NULL) return(0);
  if (di->cwlen == d->len) return(0);

  rval = fwrite(d->data + di->cwlen, 1, d->len - di->cwlen, di->fp);
  if (rval > 0) di->cwlen += (size_t)rval;
  else ; /* do something about this */

  return(0);
}

/*
 * DownloadEndData
 */
static int
DownloadEndData(di, d)
DownloadInfo *di;
Document *d;
{
  int rval;

  di->finished = true;
  di->d = d;

  if (di->fp == NULL) return(0);

  if (di->cwlen < d->len)
  {
    rval = fwrite(d->data + di->cwlen, d->len - di->cwlen, 1, di->fp);
    if (rval > 0) di->cwlen += (size_t)rval;
    else ; /* do something about this */
  }

  fclose(di->fp);
  di->fp = NULL;

  return(0);
}

/*
 * DownloadDestroy
 */
static int
DownloadDestroy(di)
DownloadInfo *di;
{
  if (di->fp != NULL) fclose(di->fp);
  if (di->w != NULL) XtDestroyWidget(di->w);
  if (di->save != NULL) XtDestroyWidget(di->save);
  MPDestroy(di->mp);
  return(0);
}

/*
 * DownloadExpose
 */
static int
DownloadExpose(di, x, y, width, height)
DownloadInfo *di;
int x, y;
unsigned int width, height;
{
  return(0);
}

/*
 * DownloadInit
 */
static void *
DownloadInit(up, d, class_context, frame)
URLParts *up;
Document *d;
void *class_context;
Frame frame;
{
  DownloadInfo *di;
  unsigned int width, height;
  unsigned int twidth, theight;
  unsigned int swidth;
  XFontStruct *font;
  Widget fw = (Widget)FrameDrawWidget(frame);
  char *label;
  int tx;
  MemPool mp;

  mp = MPCreate();
  di = (DownloadInfo *)MPCGet(mp, sizeof(DownloadInfo));
  di->mp = mp;
  di->frame = frame;
  di->up = URLDup(mp, up);
  di->finished = false;

  if (FrameGetSize(frame, true, &width, &height) == -1)
  {
    width = 200;
    height = 200;
    FrameReturnSize(frame, true, width, height);
  }

  di->w = XtVaCreateWidget("download_text",
			   scrollingTextWidgetClass, fw,
			   NULL);

  XtVaGetValues(di->w, XtNtextWidget, &(di->tw), NULL);
  XtVaGetValues(di->tw, XtNfont, &font, NULL);
  if (font == NULL) theight = 50;
  else theight = font->ascent + font->descent + 6;
  twidth = XTextWidth(font, "X", 1) * 30;

  di->save = XtVaCreateWidget("download_save",
			      commandWidgetClass, fw,
			      NULL);
  XtVaGetValues(di->save, XtNfont, &font, XtNlabel, &label, NULL);
  if (label == NULL) label = "download_save";
  swidth = XTextWidth(font, label, strlen(label));

  XtVaSetValues(di->tw,
		XtNeditType, XtEtextEdit,
		XtNstring, up->filename,
		XtNecho, True,
		NULL);

  tx = width / 2 - (twidth + swidth + 5) / 2;
  XtVaSetValues(di->w,
		XtNx, tx,
		XtNy, height / 2 - theight / 2,
		XtNwidth, twidth,
		XtNheight, theight,
		NULL);

  XtVaSetValues(di->save,
		XtNx, tx + twidth + 5,
		XtNy, height / 2 - theight / 2,
		NULL);

  XtManageChild(di->w);
  XtManageChild(di->save);

  XtAddCallback(di->save, XtNcallback, DownloadSaveCallback, (XtPointer)di);

  return(di);
}

void
InitModule_Download(wc)
WWWContext wc;
{
  static RenderHooks rh;

  memset(&rh, 0, sizeof(RenderHooks));
  rh.content = "x-www/x-download";
  rh.init = DownloadInit;
  rh.add = DownloadAddData;
  rh.end = DownloadEndData;
  rh.expose = DownloadExpose;
  rh.destroy = DownloadDestroy;
  WWWAddRenderHooks(wc, &rh);

  return;
}

