/*
 * file.c
 *
 * Copyright (c) 1995, John Kilburg <john@cs.unlv.edu>
 * See the file COPYRIGHT for details.
 */
#include <stdio.h>
#include <ctype.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>

#include "WWW.h"

/*
 * This is added to _POSIX_PATH_MAX with the result used as the size of a
 * generic buffer used to build strings and directory entries.  You might
 * have to make it bigger if the hardcoded messages in DirRead() get longer.
 * It should also be at least twice as long as the string representation of
 * the largest integer.
 */
#define MUMBLE 1024

#if defined(SYSV) || defined(SVR4) || defined(__arm)
#include <dirent.h>
#define DIRSTUFF struct dirent
#else
#include <sys/dir.h>
#define DIRSTUFF struct direct
#endif

/* Jim Rees fix */
#ifndef S_ISDIR
#define S_ISDIR(mode) (((mode) & S_IFMT) == S_IFDIR)
#endif

#ifndef S_ISBLK
#define S_ISBLK(mode) (((mode) & S_IFMT) == S_IFBLK)
#endif

#ifndef S_ISREG
#define S_ISREG(mode) (((mode) & S_IFMT) == S_IFREG)
#endif

#ifndef S_ISFIFO
#define S_ISFIFO(mode) (((mode) & S_IFMT) == S_IFIFO)
#endif

#ifndef S_ISCHR
#define S_ISCHR(mode) (((mode) & S_IFMT) == S_IFCHR)
#endif

/* not POSIX but what the heck */
#ifndef S_ISLNK
#define S_ISLNK(mode) (((mode) & S_IFMT) == S_IFLNK)
#endif

typedef struct _reg_info
{
  FILE *fp;
  off_t size;
} RegInfo;

typedef struct _dir_info
{
  DIR *dp;
  int size;
  int used;
  char *dirname;
  char **sa;
  char *direntry;
} DirInfo;

typedef struct _file_info
{
  MemPool mp;
  bool directory;
  Frame frame;
  URLParts *up;
  IOTimeOut iot;
  DirInfo di;
  RegInfo ri;
  Document *d;
  bool init_done;
  char *mbuf;
  char *rstr;
} FileInfo;

static int
local_strcmp(a, b)
char **a, **b;
{
  return(strcmp(*a, *b));
}

/*
 * DirEOF
 */
static void
DirEOF(fi)
FileInfo *fi;
{
  int i;
  char *f;
  char **sa = fi->di.sa;
  char *cp;
  int flen;
  static char *header1 = "";
  static char *header2 = "<ul>\n";
  static char *trailer1 = "\n</ul>";
  static char *format = "<li><a href=file:%s%s>%s</a>%s\n";

  if (fi->di.used > 0) qsort(sa, fi->di.used, sizeof(char *), local_strcmp);

  for (i = 0, flen = 0; i < fi->di.used; i++)
  {
    flen += 2 * strlen(sa[i]) + strlen(format) + strlen(fi->di.dirname);
  }
  flen += strlen(header1) + strlen(header2) + strlen(trailer1) + sizeof(char);

  f = (char *)alloc_mem(flen);
  strcpy(f, header1);
  strcat(f, header2);
  for (i = 0; i < fi->di.used; i++)
  {
    for (cp = sa[i]; *cp != '\0'; cp++)
    {
      if (isspace(*cp)) break;
    }
    *cp++ = '\0';
    sprintf(f + strlen(f), format, fi->di.dirname, sa[i], sa[i], cp); 
  }
  strcat(f, trailer1);

  fi->d->ds = fi->d->data = (byte *)f;
  fi->d->len = strlen(f);
  fi->init_done = true;
  FrameInitData(fi->frame, fi->d);
  FrameEndData(fi->frame, fi->d);

  return;
}

/*
 * DirRead
 */
static void
DirRead(iot, fi)
IOTimeOut iot;
FileInfo *fi;
{
  DIRSTUFF *de;
  int salen;
  struct stat fs;
  char *buf = fi->mbuf;

  if ((de = readdir(fi->di.dp)) == NULL)
  {
    closedir(fi->di.dp);
    fi->di.dp = NULL;
    DirEOF(fi);
    return;
  }

  /* Note: use of buf changes after stat() */
  strcpy(buf, fi->di.dirname);
  strcat(buf, de->d_name);
  if (stat(buf, &fs) != -1)
  {
    if (S_ISREG(fs.st_mode)) sprintf(buf, " (%ld bytes)", (long)fs.st_size);
    else if (S_ISDIR(fs.st_mode)) strcpy(buf, "/");
    else if (S_ISLNK(fs.st_mode)) strcpy(buf, " &lt;link&gt;");
    else if (S_ISFIFO(fs.st_mode)) strcpy(buf, " &lt;pipe&gt;");
    else if (S_ISBLK(fs.st_mode)) strcpy(buf, " &lt;block device&gt;");
    else if (S_ISCHR(fs.st_mode)) strcpy(buf, " &lt;char device&gt;");
    else strcpy(buf, " &lt;unknown&gt;");
  }
  else strcpy(buf, " ");
  
  /*
   * Resize the file entry table if needed.
   */
  if (fi->di.used >= fi->di.size)
  {
    char **nsa;
    nsa = (char **)MPGet(fi->mp, fi->di.size * 2 * sizeof(char *));
    memcpy(nsa, fi->di.sa, fi->di.size * sizeof(char *));
    fi->di.size *= 2;
  }
  
  salen = strlen(de->d_name) + strlen(buf) + 2 * sizeof(char);
  fi->di.sa[fi->di.used] = (char *)MemPoolGet(fi->mp, salen);
  strcpy(fi->di.sa[fi->di.used], de->d_name);
  strcat(fi->di.sa[fi->di.used], " ");
  strcat(fi->di.sa[fi->di.used], buf);
  fi->di.used++;

  sprintf (fi->mbuf, fi->rstr, fi->di.used, fi->up->filename);
  FrameMessage(fi->frame, fi->mbuf);

  IOSetTimeOut(fi->iot, 10, DirRead);

  return;
}

/*
 * FileRead
 */
static void
FileRead(iot, fi)
IOTimeOut iot;
FileInfo *fi;
{
  int readlen;

  readlen = fi->ri.size - fi->d->len;
  if (readlen > BUFSIZ) readlen = BUFSIZ;

  if ((readlen = fread(fi->d->data + fi->d->len,
		       1, readlen, fi->ri.fp)) == 0)
  {
    fclose(fi->ri.fp);
    fi->ri.fp = NULL;
    if (fi->init_done) FrameEndData(fi->frame, fi->d);
  }
  else
  {
    if (fi->d->len == 0)
    {
      fi->init_done = true;
      FrameInitData(fi->frame, fi->d);
    }
    fi->d->len += readlen;
    FrameAddData(fi->frame, fi->d);

    sprintf (fi->mbuf, fi->rstr, fi->d->len, fi->ri.size, fi->up->filename);
    FrameMessage(fi->frame, fi->mbuf);

    IOSetTimeOut(fi->iot, 10, FileRead);
  }

  return;
}

/*
 * FileDestroy
 */
static int
FileDestroy(fi)
FileInfo *fi;
{
  if (fi->di.dp != NULL) closedir(fi->di.dp);
  if (fi->ri.fp != NULL) fclose(fi->ri.fp);
  if (!fi->init_done && fi->d != NULL) DocumentDestroy(fi->d);
  if (fi->iot != NULL) IODestroyTimeOut(fi->iot);
  MPDestroy(fi->mp);

  return(0);
}

/*
 * FileInit
 */
static void *
FileInit(up, pup, class_context, frame)
URLParts *up;
URLParts *pup;
void *class_context;
Frame frame;
{
  FileInfo *fi;
  struct stat s;
  void (*func)();
  char *rname;
  char *content;
  MemPool mp;

  if (stat(up->filename, &s) == -1) return(NULL);

  mp = MPCreate();
  fi = (FileInfo *)MPCGet(mp, sizeof(FileInfo));
  fi->mp = mp;
  fi->frame = frame;
  fi->up = URLDup(fi->mp, up);
  fi->iot = IOCreateTimeOut(frame, fi);
  fi->d = DocumentCreate(true);

  if (S_ISDIR(s.st_mode))
  {
    fi->directory = true;
    if ((fi->di.dp = opendir(up->filename)) == NULL)
    {
      FileDestroy(fi);
      return(NULL);
    }

    fi->di.dirname = (char *)MPGet(fi->mp, strlen(up->filename) + 
				   strlen("/") + sizeof(char));
    strcpy(fi->di.dirname, up->filename);
    if (up->filename[strlen(up->filename) - 1] != '/')
    {
      strcat(fi->di.dirname, "/");
    }

    fi->di.size = 512;
    fi->di.sa = (char **)MPGet(fi->mp, sizeof(char **) * fi->di.size);

    rname = "file.readdir";
    func = DirRead;
    content = "text/html";
  }
  else
  {
    fi->directory = false;
    if ((fi->ri.fp = fopen(up->filename, "r")) == NULL)
    {
      FileDestroy(fi);
      return(NULL);
    }

    fi->d->ds = fi->d->data = (byte *)alloc_mem((int)s.st_size);
    fi->ri.size = s.st_size;

    rname = "file.readfile";
    func = FileRead;
    if ((content = MIMEExt2Content(up->filename)) == NULL)
    {
      content = "text/plain";
    }
  }

  fi->d->mflist = MIMECreateField(fi->d->mp, "content-type", content);
  fi->rstr = FrameGetResource(frame, rname);
  if (fi->rstr == NULL) fi->rstr = rname;

  fi->mbuf = (char *)MemPoolGet(fi->mp, strlen(fi->rstr) +
				_POSIX_PATH_MAX + MUMBLE);

  IOSetTimeOut(fi->iot, 10, func);

  return(fi);
}

/*
 * InitModule_File
 */
void
InitModule_File(wc)
WWWContext wc;
{
  static ProtoHooks ph;

  ph.protocol = "file";
  ph.init = FileInit;
  ph.destroy = FileDestroy;
  WWWAddProtoHooks(wc, &ph);

  return;
}

