/*
 * ftp.c
 *
 * Copyright 1993, 1994, 1995 by John Kilburg (john@cs.unlv.edu)
 * See the file COPYRIGHT for details.
 */
#include <stdio.h>
#include <ctype.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include "WWW.h"

/*
 * Should create a FTP Class which keeps track of connections
 * between URLs and times them out.
 */

typedef struct _ftpinfo
{
  Frame frame;
  IOStream ds;
  IOStream cs;
  Document *d;
  void (*rfunc)();
  URLParts *up;
  MemPool mp;
  int size;
  bool ignore_err;
} FTPInfo;

static char *FTPDirToHTML _ArgProto((FTPInfo *, byte *, int));
static void FTPSimple _ArgProto((FTPInfo *, char *, void (*)(), bool));

/*
 * FTPDestroy
 */
static int
FTPDestroy(fi)
FTPInfo *fi;
{
  if (fi->ds != NULL) IODestroyStream(fi->ds);
  if (fi->cs != NULL) IODestroyStream(fi->cs);
  MPDestroy(fi->mp);

  return(0);
}

/*
 * FTPData
 */
static void
FTPData(ios, eof, closure)
IOStream ios;
bool eof;
void *closure;
{  
  FTPInfo *fi = (FTPInfo *)closure;

  fi->d->data = fi->d->ds = IOStreamBuffer(ios);
  fi->d->len = IOStreamBufferLength(ios);

  if (eof)
  {
    IOSnarfStreamBuffer(ios);
    IODestroyStream(fi->ds);
    fi->ds = NULL;
    IODestroyStream(fi->cs);
    fi->cs = NULL;
    FrameEndData(fi->frame, fi->d);
  }
  else
  {
    FrameAddData(fi->frame, fi->d);
    IOReadStream(ios, FTPData);
  }

  return;
}

/*
 * ftp_dirdata
 */
static void
FTPDirData(ios, eof, fi)
IOStream ios;
bool eof;
FTPInfo *fi;
{  
  if (eof)
  {
    MIMEField *mf;

    fi->d = DocumentCreate(false);
    mf = MIMECreateField(fi->d->mp, "content-type", "text/html");
    mf->next = fi->d->mflist;
    fi->d->mflist = mf;
    fi->d->data = fi->d->ds = FTPDirToHTML(fi,
					   IOStreamBuffer(ios),
					   IOStreamBufferLength(ios));
    fi->d->len = strlen(fi->d->data);
    IODestroyStream(fi->cs);
    fi->cs = NULL;
    IODestroyStream(fi->ds);
    fi->ds = NULL;
    FrameInitData(fi->frame, fi->d);
    FrameEndData(fi->frame, fi->d);
  }
  else IOReadStream(fi->ds, FTPDirData);

  return;
}

/*
 * FTPParseResponse
 */
static int
FTPParseResponse(data, len)
byte *data;
int len;
{
  byte *cp;
  
  if (data[len - 1] == '\n')
  {
    if (*(data + 3) == ' ') return(atoi(data));
    for (cp = data + len - 2; cp >= data; cp--)
    {
      if (*cp == '\n' && *(cp + 4) == ' ') return(atoi(cp + 1));
    }
  }

  return(-1);
}

/*
 * FTPSimpleRead
 */
static void
FTPSimpleRead(ios, eof, fi)
IOStream ios;
bool eof;
FTPInfo *fi;
{
  int ecode;
  byte *data;
  int len;

  data = IOStreamBuffer(ios);
  len = IOStreamBufferLength(ios);
  if ((ecode = FTPParseResponse(data, len)) == -1)
  {
    IOReadStream(fi->cs, FTPSimpleRead);
    return;
  }
  if (ecode < 400 || fi->ignore_err) (fi->rfunc)(fi, data);
/*  else ftp_error_return(fi, data); */

  return;
}

/*
 * FTPWrite
 */
static void
FTPWrite(ios, closure)
IOStream ios;
void *closure;
{
  FTPInfo *fi = (FTPInfo *)closure;

  IOResetStream(fi->cs);
  IOReadStream(fi->cs, FTPSimpleRead);
  return;
}

/*
 * FTPSimple
 */
static void
FTPSimple(fi, request, rfunc, ignore_err)
FTPInfo *fi;
char *request;
void (*rfunc)();
bool ignore_err;
{
  fi->rfunc = rfunc;
  fi->ignore_err = ignore_err;
  IOWriteStream(fi->cs, request, strlen(request), FTPWrite);
  return;
}

/*
 * FTPDirRead
 */
static void
FTPDirRead(fi, data)
FTPInfo *fi;
byte *data;
{
  IOResetStream(fi->ds);
  IOReadStream(fi->ds, FTPDirData);
  return;
}

/*
 * FTPNlst
 */
static void
FTPNlst(fi, data)
FTPInfo *fi;
byte *data;
{
  FTPSimple(fi, "NLST\r\n", FTPDirRead, false);
  return;
}

/*
 * FTPDummy
 */
static void
FTPDummy(ios, eof, fi)
IOStream ios;
bool eof;
FTPInfo *fi;
{
  return;
}

/*
 * FTPCwd
 */
static void
FTPCwd(fi, data)
FTPInfo *fi;
byte *data;
{
  int ecode;

  sscanf(data, "%d", &ecode);
  if (ecode < 400)
  {
    char *content;
    MIMEField *mf;
    
    fi->d = DocumentCreate(false);
    content = MIMEExt2Content(fi->up->filename);
    if (content == NULL) content = "text/plain";
    mf = MIMECreateField(fi->d->mp, "content-type", content);
    mf->next = fi->d->mflist;
    fi->d->mflist = mf;

    if (fi->size > 0) IOSetStreamBuffer(fi->ds, alloc_mem(fi->size), fi->size);
    IOReadStream(fi->ds, FTPData);

    IOResetStream(fi->cs);
    IOReadStream(fi->cs, FTPDummy);

    FrameInitData(fi->frame, fi->d);
  }
  else
  {
    char *format = "CWD %s\r\n";
    char *request;

    request = MemPoolGet(fi->mp,
			 strlen(format) + strlen(fi->up->filename) + 1);
    sprintf (request, format, fi->up->filename);
    FTPSimple(fi, request, FTPNlst, false);
  }

  return;
}

/*
 * FTPRetrieve
 */
static void
FTPRetrieve(fi, data)
FTPInfo *fi;
byte *data;
{
  int h0, h1, h2, h3, p0, p1, reply, n;
  char *format = "RETR %s\r\n";
  char *request;
  char dhost[BUFSIZ];
  int dport;

  n = sscanf((char *)data, "%d %*[^(] (%d,%d,%d,%d,%d,%d)",
	     &reply, &h0, &h1, &h2, &h3, &p0, &p1);
  if (n != 7 || reply != 227)
  {
    /* error */
    return;
  }
  
  sprintf (dhost, "%d.%d.%d.%d", h0, h1, h2, h3);
  dport = (p0 << 8) + p1;
  
  /*
   * Check for error here.
   */
  if ((fi->ds = IOCreateStream(fi->frame, dhost, dport, NULL, fi)) == NULL)
  {
    return;
  }

  request = MemPoolGet(fi->mp, strlen(format) + strlen(fi->up->filename) + 1);
  sprintf (request, format, fi->up->filename);
  FTPSimple(fi, request, FTPCwd, true);

  return;
}

/*
 * FTPPassive
 */
static void
FTPPassive(fi, data)
FTPInfo *fi;
byte *data;
{
  int ecode;

  sscanf(data, "%d %d", &ecode, &(fi->size));
  if (ecode >= 400) fi->size = 0;
  FTPSimple(fi, "PASV\r\n", FTPRetrieve, true);

  return;
}

/*
 * FTPSize
 */
static void
FTPSize(fi, data)
FTPInfo *fi;
byte *data;
{
  char *request;
  char *format = "SIZE %s\r\n";
  
  request = MemPoolGet(fi->mp, strlen(format) + strlen(fi->up->filename) + 1);
  sprintf (request, format, fi->up->filename);
  FTPSimple(fi, request, FTPPassive, true);
  return;
}

/*
 * FTPType
 */
static void
FTPType(fi, data)
FTPInfo *fi;
byte *data;
{
  FTPSimple(fi, "TYPE I\r\n", FTPSize, false);
  return;
}

/*
 * FTPPass
 */
static void
FTPPass(fi, data)
FTPInfo *fi;
byte *data;
{
  char *request;
  char *uname;

  if ((uname = getenv("EMAIL")) != NULL)
  {
    request = MemPoolGet(fi->mp, strlen(uname) + 9);
    strcpy(request, "PASS -");
    strcat(request, uname);
    strcat(request, "\r\n");
  }
  else
  {
    request = MemPoolStrDup(fi->mp, "PASS -chimera@nowhere.org\r\n");
  } 
  FTPSimple(fi, request, FTPType, false);

  return;
}

/*
 * FTPUser
 */
static void
FTPUser(fi, data)
FTPInfo *fi;
byte *data;
{
  FTPSimple(fi, "USER anonymous\r\n", FTPPass, false);
  return;
}

/*
 * ProtoInit_FTP
 */
static void *
FTPInit(up, pup, class_context, frame)
URLParts *up;
URLParts *pup;
void *class_context;
Frame frame;
{
  FTPInfo *fi;
  URLParts *nup;
  MemPool mp;

  mp = MPCreate();

  if (pup != NULL)
  {
    if ((nup = URLParse(mp, "x-message:noftpproxy")) != NULL)
    {
      FrameLoad(frame, nup);
      MPDestroy(mp);
    }    
  }

  fi = (FTPInfo *)MPCGet(mp, sizeof(FTPInfo));
  fi->mp = mp;
  fi->cs = IOCreateStream(frame, up->hostname,
			  up->port == 0 ? 21:up->port, NULL, fi);
  if (fi->cs == NULL)
  {
    if ((nup = URLParse(fi->mp, "x-message:openfailed")) != NULL)
    {
      FrameLoad(frame, nup);
    }
    FTPDestroy(fi);
    return(NULL);
  }

  fi->frame = frame;
  fi->up = URLDup(fi->mp, up);

  fi->rfunc = FTPUser;
  IOResetStream(fi->cs);
  IOReadStream(fi->cs, FTPSimpleRead);

  return(fi);
}

/*
 * InitModule_FTP
 */
void
InitModule_FTP(wc)
WWWContext wc;
{
  static ProtoHooks ph;

  memset(&ph, 0, sizeof(ProtoHooks));
  ph.protocol = "ftp";
  ph.init = FTPInit;
  ph.destroy = FTPDestroy;
  WWWAddProtoHooks(wc, &ph);
  
  return;
}

static int
ftp_strcmp(a, b)
char **a, **b;
{
  return(strcmp(*a, *b));
}

/*
 * FTPDirToHTML
 *
 * Convert directory list to HTML
 */
static char *
FTPDirToHTML(fi, data, len)
FTPInfo *fi;
byte *data;
int len;
{
  char *f;
  int i;
  char *hostname;
  char *filename;
  char *header;
  char *trailer;
  char *entry = "<li><a href=ftp://%s:%d%s/%s>%s</a>\n";
  byte *cp, *lastcp, *dname;
  int sacount;
  char **sa;
  int olen, hlen, flen, elen;

  header = FrameGetResource(fi->frame, "ftp.dirheader");
  if (header == NULL) header = "<html><body><h2>FTP Directory</h2>";

  trailer = FrameGetResource(fi->frame, "ftp.dirtrailer");
  if (trailer == NULL) trailer = "</body></html>";

  hostname = fi->up->hostname;
  filename = fi->up->filename;
  hlen = strlen(hostname);
  flen = strlen(filename);
  elen = strlen(entry);

  sacount = 0;
  for (cp = data, lastcp = cp + len; cp < lastcp; cp++)
  {
    if (*cp == '\n') sacount++;
  }
  sa = (char **)MemPoolGet(fi->mp, sizeof(char *) * sacount);
  dname = data;
  olen = 0;
  for (i = 0, cp = data, lastcp = cp + len; cp < lastcp; cp++)
  {
    if (*cp == '\n')
    {
      sa[i] = dname;
      *cp = '\0';
      dname = cp + 1;
      olen += hlen + 20 + flen + elen + strlen(sa[i]) * 2;
      i++;
    }
  }
  qsort(sa, sacount, sizeof(char *), ftp_strcmp);

  olen += strlen(header) + 2 * flen + hlen + strlen(trailer) + 1;
  f = (char *)alloc_mem(olen);
  sprintf (f, header, filename, hostname, filename);

  if (filename[0] != '\0' && filename[1] == '\0') filename = "";

  for (i = 0; i < sacount; i++)
  {
    sprintf(f + strlen(f), entry,
	    hostname,
	    fi->up->port == 0 ? 21:fi->up->port,
	    filename,
	    sa[i], sa[i]);
  }

  strcat(f, trailer);

  return(f);
}

