/*
 * http.c
 *
 * Copyright (C) 1993, 1994, 1995, John Kilburg <john@cs.unlv.edu>
 * See the file COPYRIGHT for details.
 */
#include <stdio.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include "WWW.h"

typedef struct _httpinfo
{
  MemPool mp;
  Frame frame;
  URLParts *up;
  URLParts *pup;
  byte *request;
  IOStream ios;
  int dstart;
  Document *d;
} HTTPInfo;

static char *HTTPBuildAuthInfo _ArgProto((HTTPInfo *, URLParts *));
static char *HTTPBuildRequest _ArgProto((HTTPInfo *));

static char *format_get = "\
GET %s HTTP/1.0\r\n\
User-Agent: %s\r\n\
Accept: */*\r\n\
%s\
%s\
\r\n";

static char *format_get2 = "\
GET %s?%s HTTP/1.0\r\n\
User-Agent: %s\r\n\
Accept: */*\r\n\
%s\
%s\
\r\n";

static char *format_post = "\
POST %s HTTP/1.0\r\n\
User-Agent: %s\n\
Accept: */*\r\n\
Content-length: %d\r\n\
Content-type: %s\r\n\
%s\
%s\
\r\n";

/*
 * crap for uuencode
 */
static char six2pr[64] =
{
    'A','B','C','D','E','F','G','H','I','J','K','L','M',
    'N','O','P','Q','R','S','T','U','V','W','X','Y','Z',
    'a','b','c','d','e','f','g','h','i','j','k','l','m',
    'n','o','p','q','r','s','t','u','v','w','x','y','z',
    '0','1','2','3','4','5','6','7','8','9','+','/'
};

/*
 * uuencode
 *
 * I snarfed this code from some version of libwww.
 *
 * ACKNOWLEDGEMENT:
 *      This code is taken from rpem distribution, and was originally
 *      written by Mark Riordan.
 *
 * AUTHORS:
 *      MR      Mark Riordan    riordanmr@clvax1.cl.msu.edu
 *      AL      Ari Luotonen    luotonen@dxcern.cern.ch
 *
 */
static int
uuencode(bufin, nbytes, bufcoded)
unsigned char *bufin;
unsigned int nbytes;
char *bufcoded;
{
  /* ENC is the basic 1 character encoding function to make a char printing */
#define ENC(c) six2pr[c]
  
  register char *outptr = bufcoded;
  unsigned int i;
  
  for (i=0; i<nbytes; i += 3) 
  {
    *(outptr++) = ENC(*bufin >> 2);            /* c1 */
    *(outptr++) = ENC(((*bufin << 4) & 060) | ((bufin[1] >> 4) & 017)); /*c2*/
    *(outptr++) = ENC(((bufin[1] << 2) & 074) | ((bufin[2] >> 6) & 03));/*c3*/
    *(outptr++) = ENC(bufin[2] & 077);         /* c4 */
    
    bufin += 3;
  }
  
  /* If nbytes was not a multiple of 3, then we have encoded too
   * many characters.  Adjust appropriately.
   */
  if (i == nbytes + 1)
  {
    /* There were only 2 bytes in that last group */
    outptr[-1] = '=';
  }
  else if (i == nbytes + 2)
  {
    /* There was only 1 byte in that last group */
    outptr[-1] = '=';
    outptr[-2] = '=';
  }
  *outptr = '\0';

  return(outptr - bufcoded);
}

/*
 * HTTPBuildAuthInfo
 *
 * Build the string needed for authorization.
 */
static char *
HTTPBuildAuthInfo(hi, up)
HTTPInfo *hi;
URLParts *up;
{
  char *garbage = NULL;
  char *auth_info = NULL;
  static char *format = "Authorization: %s %s\n";

  garbage = (char *)MPGet(hi->mp, strlen(up->username) +
			  1 +
			  strlen(up->password != NULL ? up->password:"") +
			  1);
  strcpy(garbage, up->username);
  if (up->password != NULL)
  {
    strcat(garbage, ":");
    strcat(garbage, up->password);
  }

  if (up->auth_type == NULL) up->auth_type = "Basic";
  if (strcmp(up->auth_type, "Basic") == 0)
  {
    char *t;

    t = (char *)MPGet(hi->mp, strlen(garbage) * 2);
    uuencode(garbage, strlen(garbage), t);
    garbage = t;
  }

  auth_info = (char *)MPGet(hi->mp, strlen(format) +
			    strlen(up->auth_type) +
			    strlen(garbage) + 1);
  sprintf (auth_info, format,
	   up->auth_type,
	   garbage);

  return(auth_info);
}

/* 
 * HTTPBuildRequest
 */
static char *
HTTPBuildRequest(hi)
HTTPInfo *hi;
{
  char *query = NULL;
  char *auth_info = NULL;
  char *filename;
  char *extra_header;
  URLParts *up;
  char *user_agent;

  user_agent = FrameGetResource(hi->frame, "http.user_agent");
  if (user_agent == NULL) user_agent = "JDK-HTTP/1.0";

  if (hi->pup != NULL) up = hi->pup;
  else up = hi->up;

  if (hi->pup != NULL) filename = URLMakeString(hi->mp, up, 0);
  else filename = MPStrDup(hi->mp, up->filename);

  if (up->reload) extra_header = "Pragma: no-cache\r\n";
  else extra_header = "";

  if (up->auth_type != NULL && up->username != NULL)
  {
    auth_info = HTTPBuildAuthInfo(hi, up);
  }
  if (auth_info == NULL) auth_info = MPStrDup(hi->mp, "");

  if (up->input_data != NULL)
  {
    if (up->method == NULL || strcasecmp(up->method, "GET") == 0)
    {
      query = (char *)MPGet(hi->mp, strlen(filename) +
			    strlen(format_get2) +
			    strlen(up->input_data) +
			    strlen(extra_header) +
			    strlen(auth_info) + 
			    strlen(user_agent) + 1);
      sprintf (query,
	       format_get2,
	       filename,
	       up->input_data,
	       user_agent,
	       extra_header,
	       auth_info);
    }
    else if (strcasecmp(up->method, "POST") == 0) /* POST ! */
    {
      query = (char *)MPGet(hi->mp, strlen(filename) +
			    strlen(format_post) +
			    up->input_data_len +
			    strlen(up->input_data_type) +
			    strlen(extra_header) +
			    strlen(auth_info) +
			    strlen(user_agent) + 1);
      sprintf (query,
	       format_post,
	       filename,
	       user_agent,
	       up->input_data_len,
	       up->input_data_type,
	       extra_header,
	       auth_info);
      memcpy(query + strlen(query), up->input_data, up->input_data_len);
    }
    else query = NULL;
  }
  else /* regular GET */
  {
    query = (char *)MPGet(hi->mp, strlen(filename) +
			  strlen(format_get) +
			  strlen(extra_header) +
			  strlen(auth_info) + 
			  strlen(user_agent) + 1);
    sprintf (query,
	     format_get,
	     filename,
	     user_agent,
	     extra_header,
	     auth_info);
  }

  return(query);
}

/*
 * HTTPParseHeader
 *
 * This expects the entire header with the line
 *
 * HTTP/x.x num mumble
 *
 * at the top.
 */
static void
HTTPParseHeader(hi, data, len)
HTTPInfo *hi;
char *data;
int len;
{
  char *cp;
  int i;

  for (i = 0, cp = data + 9; i < len; i++, cp++)
  {
    if (*cp == '\n') break;
  }

  if (i == len) return;

  hi->d = DocumentCreate(false);
  hi->d->mflist = MIMEParseHeader(hi->d->mp, cp + 1, len - (cp - data));

  FrameInitData(hi->frame, hi->d);

  return;
}

/*
 * HTTPReadData
 */
static void
HTTPReadData(ios, eof, hi)
IOStream ios;
bool eof;
HTTPInfo *hi;
{
  Document *d = hi->d;

  d->ds = IOStreamBuffer(ios);
  d->data = d->ds + hi->dstart;
  d->len = IOStreamBufferLength(ios) - hi->dstart;

  if (eof)
  {
    IOSnarfStreamBuffer(ios);
    IODestroyStream(ios);
    hi->ios = NULL;
    FrameEndData(hi->frame, d);
  }
  else
  {
    FrameAddData(hi->frame, d);
    IOReadStream(ios, HTTPReadData);
  }

  return;
}

/*
 * HTTPReadHeader
 */
static void
HTTPReadHeader(ios, eof, hi)
IOStream ios;
bool eof;
HTTPInfo *hi;
{
  byte *data;
  byte *buffer;
  char *value;
  int len;
  int dstart;
  int blen;

  data = IOStreamBuffer(hi->ios);
  len = IOStreamBufferLength(hi->ios);

  if ((dstart = MIMEFindData(data, len)) != -1)
  {
    hi->dstart = dstart;
    HTTPParseHeader(hi, data, dstart);
    if ((value = MIMEGetField(hi->d->mflist, "location")) != NULL)
    {
      URLParts *up;
      
      if ((up = URLParse(hi->mp, value)) != NULL)
      {
	/*
         * FrameReplace will destroy the current HTTP context so return
	 * immediately and do not touch 'hi' again.
	 */
	FrameReplace(hi->frame, up);
	return;
      }
    }
    else if ((value = MIMEGetField(hi->d->mflist, "content-length")) != NULL)
    {
      blen = atoi(value);
      buffer = (char *)alloc_mem(blen + len);
      memcpy(buffer, data, len);
      IOSetStreamBuffer(hi->ios, buffer, blen + dstart);
    }

    HTTPReadData(ios, eof, hi);
  }
  else
  {
    if (!eof) IOReadStream(ios, HTTPReadHeader);
  }

  return;
}

/*
 * HTTPFakeHeader
 */
static void
HTTPFakeHeader(hi)
HTTPInfo *hi;
{
  char *content;

  hi->d = DocumentCreate(false);
  if ((content = MIMEExt2Content(hi->up->filename)) == NULL)
  {
    content = "text/plain";
  }
  hi->d->mflist = MIMECreateField(hi->d->mp, "content-type", content);
  FrameInitData(hi->frame, hi->d);
  return;
}

/*
 * HTTPRead
 */
static void
HTTPRead(ios, eof, hi)
IOStream ios;
bool eof;
HTTPInfo *hi;
{
  char *data;
  
  data = IOStreamBuffer(hi->ios);
  if (IOStreamBufferLength(hi->ios) >= 8)
  {
    if (strncmp("HTTP/1.0", data, 8) == 0) HTTPReadHeader(ios, eof, hi);
    else
    {
      HTTPFakeHeader(hi);
      HTTPReadData(ios, eof, hi);
    }
  }
  else
  {
    if (eof)
    {
      HTTPFakeHeader(hi);
      HTTPReadData(ios, true, hi);
    }
    else IOReadStream(hi->ios, HTTPRead);
  }

  return;
}

/*
 * HTTPRequestDone
 */
static void
HTTPRequestDone(ios, hi)
IOStream ios;
HTTPInfo *hi;
{
  IOResetStream(hi->ios);
  IOReadStream(hi->ios, HTTPRead);

  return;
}

/*
 * HTTPDestroy
 */
static int
HTTPDestroy(hi)
HTTPInfo *hi;
{
  if (hi->ios != NULL)
  {
    if (hi->d != NULL) IOSnarfStreamBuffer(hi->ios);
    IODestroyStream(hi->ios);
  }
  MPDestroy(hi->mp);
  return(0);
}

/*
 * HTTPInit
 */
static void *
HTTPInit(up, pup, class_context, frame)
URLParts *up;
URLParts *pup;
void *class_context;
Frame frame;
{
  HTTPInfo *hi;
  URLParts *nup;
  MemPool mp;

  mp = MPCreate();
  hi = (HTTPInfo *)MPCGet(mp, sizeof(HTTPInfo));
  hi->mp = mp;
  hi->ios = IOCreateStream(frame, up->hostname,
			   up->port == 0 ? 80:up->port, NULL, hi);
  if (hi->ios == NULL)
  {
    if ((nup = URLParse(hi->mp, "x-message:openfailed")) != NULL)
    {
      FrameLoad(frame, nup);
    }
    HTTPDestroy(hi);
    return(NULL);
  }

  hi->frame = frame;
  if (pup != NULL) hi->pup = URLDup(hi->mp, pup);
  hi->up = URLDup(hi->mp, up);

  hi->request = HTTPBuildRequest(hi);
  IOWriteStream(hi->ios, hi->request, strlen(hi->request), HTTPRequestDone);

  return(hi);
}

/*
 * InitModule_HTTP
 */
void
InitModule_HTTP(wc)
WWWContext wc;
{
  static ProtoHooks ph;

  memset(&ph, 0, sizeof(ProtoHooks));
  ph.protocol = "http";
  ph.init = HTTPInit;
  ph.destroy = HTTPDestroy;
  WWWAddProtoHooks(wc, &ph);

  return;
}


