/*
 * WWW.h
 *
 * Copyright (c) 1995, John D. Kilburg <john@cs.unlv.edu>
 * See the file COPYRIGHT for details.
 */
#ifndef __WWW_H__
#define __WWW_H__ 1

#include "common.h"

typedef struct _IOStream     *IOStream;
typedef struct _IOTimeOut    *IOTimeOut;
typedef struct _WWWContext   *WWWContext;
typedef struct _Frame        *Frame;

typedef struct _render_hooks
{
  char *content;
  void *class_context;
  void *(*init)();
  int (*reset)();
  int (*add)();
  int (*end)();
  int (*destroy)();
  int (*mouse)();
  int (*keyboard)();
  int (*expose)();
  int (*motion)();
  int (*class_destroy)();
  struct _render_hooks *next;
} RenderHooks;

typedef struct _proto_hooks
{
  char *protocol;
  void *class_context;
  void *(*init)();
  int (*destroy)();
  int (*class_destroy)();
  struct _proto_hooks *next;
} ProtoHooks;

typedef struct _conv_hooks
{
  char *stype;
  char *dtype;
  void *class_context;
  void *(*init)();
  int (*add)();
  int (*end)();
  int (*destroy)();
  int (*class_destroy)();
  struct _conv_hooks *next;
} ConvHooks;

/*
 * Some of these fields should not be here.  It was convienent.
 *
 * protocol://username:password@hostname:port/filename#anchor
 */
typedef struct _url_parts
{
  MemPool mp;
  char *protocol;
  char *hostname;
  int port;
  char *filename;
  char *anchor;
  char *method;
  void *input_data;
  size_t input_data_len;
  char *input_data_type;
  char *auth_type;
  char *auth_realm;
  char *username;
  char *password;
  bool reload;
} URLParts;

typedef struct _mimefield
{
  char *name;
  char *value;
  struct _mimefield *next;
} MIMEField;

typedef struct _document
{
  byte *ds;                        /* start of buffer, free this */
  byte *data;                      /* start of data */
  size_t len;                      /* length of data */
  MemPool mp;                      /* memory pool for MIME stuff */
  MIMEField *mflist;               /* pseudo MIME stuff */
  bool nocache;                    /* don't cache document ? */
} Document;

typedef struct _mimecap
{
  char *ctype;
  char *command;
  struct _mimecap *next;  
} MIMECap;

/*
 * mime.c
 */
MIMEField *MIMECreateField _ArgProto((MemPool, char *, char *));
MIMEField *MIMEParseHeader _ArgProto((MemPool, char *, size_t));
char *MIMEGetField _ArgProto((MIMEField *, char *));
char *MIMEExt2Content _ArgProto((char *));
void MIMEReadTypeFiles _ArgProto((char *));
void MIMEReadCapFiles _ArgProto((char *));
int MIMEFindData _ArgProto((char *, size_t));

/*
 * io.c
 */
IOStream IOCreateStream _ArgProto((Frame, char *, int, void (*)(), void *));
void IODestroyStream _ArgProto((IOStream));
void IOWriteStream _ArgProto((IOStream, byte *, size_t, void (*)()));
void IOReadStream _ArgProto((IOStream, void (*)()));
void IOSetStreamBuffer _ArgProto((IOStream, byte *, size_t));
byte *IOStreamBuffer _ArgProto((IOStream));
size_t IOStreamBufferLength _ArgProto((IOStream));
byte *IOSnarfStreamBuffer _ArgProto((IOStream));
IOTimeOut IOCreateTimeOut _ArgProto((Frame, void *));
void IOSetTimeOut _ArgProto((IOTimeOut, unsigned int, void (*)()));
void IODestroyTimeOut _ArgProto((IOTimeOut));
void IOResetStream _ArgProto((IOStream));

/*
 * url.c
 */
int URLcmp _ArgProto((URLParts *, URLParts *));
char *URLMakeString _ArgProto((MemPool, URLParts *, int));
URLParts *URLResolve _ArgProto((MemPool, URLParts *, URLParts *));
URLParts *URLParse _ArgProto((MemPool, char *));
char *URLUnescape _ArgProto((MemPool, char *));
char *URLEscape _ArgProto((MemPool, char *, bool));
URLParts *URLDup _ArgProto((MemPool, URLParts *));
bool URLIsAbsolute _ArgProto((URLParts *));
char *URLHash _ArgProto((MemPool, URLParts *));

/*
 * frame.c
 */
int FrameInitData _ArgProto((Frame, Document *));
int FrameAddData _ArgProto((Frame, Document *));
int FrameEndData _ArgProto((Frame, Document *));

int FrameLoad _ArgProto((Frame, URLParts *));
int FrameReload _ArgProto((Frame));
int FrameDownload _ArgProto((Frame, URLParts *));
int FrameBack _ArgProto((Frame));
int FrameReplace _ArgProto((Frame, URLParts *));
int FrameCancel _ArgProto((Frame));
int FrameHome _ArgProto((Frame));
int FrameDestroy _ArgProto((Frame));
int FrameOverride _ArgProto((Frame, RenderHooks *, void *));

Document *DocumentCreate _ArgProto((bool));
void DocumentDestroy _ArgProto((Document *));

/*
 * gui.c
 */
void FrameGetOnScreenDimensions _ArgProto((Frame, int *, int *,
					   unsigned int *, unsigned int *));
int FrameGetNamedColor _ArgProto((Frame, char *, unsigned long *));

char *FrameGetResource _ArgProto((Frame, char *));

Frame FrameCreateUnsized _ArgProto((Frame, void (*)(), void *));
Frame FrameCreateSized _ArgProto((Frame, unsigned int, unsigned int));
Frame FrameCreate _ArgProto((WWWContext, Frame));
void FrameAssignName _ArgProto((Frame, char *));
Frame FrameCreateInv _ArgProto((Frame));

int FrameMap _ArgProto((Frame, int, int));
int FrameGetSize _ArgProto((Frame, bool, unsigned int *, unsigned int *));
int FrameSetSize _ArgProto((Frame, bool, unsigned int, unsigned int));
int FrameReturnSize _ArgProto((Frame, bool, unsigned int, unsigned int));
int FrameDestroy _ArgProto((Frame));

int FrameGetAuth _ArgProto((Frame, void (*)(), void *));

int FrameMessage _ArgProto((Frame, char *));
void *FrameWidget _ArgProto((Frame));
void *FrameDrawWidget _ArgProto((Frame));

/*
 * context.c
 */
#define WWW_RENDER 0
#define WWW_PROTO 1
#define WWW_CONV 2

typedef bool (*WWWResolveProc) _ArgProto((WWWContext, void *, int, char *));
typedef void (*WWWURLProc) _ArgProto((Frame, void *, char *));
typedef void (*WWWMessageProc) _ArgProto((Frame, void *, char *));

int WWWAddProtoHooks _ArgProto((WWWContext, ProtoHooks *));
int WWWAddRenderHooks _ArgProto((WWWContext, RenderHooks *));
int WWWAddConvHooks _ArgProto((WWWContext, ConvHooks *));
char *WWWGetResource _ArgProto((WWWContext, char *));
WWWContext WWWCreateContext _ArgProto((WWWResolveProc, void *));
void WWWAddResourceFile _ArgProto((WWWContext, char *));
void WWWDestroyContext _ArgProto((WWWContext));
Frame WWWCreateToplevelFrame _ArgProto((WWWContext, void *, char *,
					WWWURLProc, WWWMessageProc,
					void *));
WWWContext FrameToWWWContext _ArgProto((Frame));
Frame WWWFindNamedFrame _ArgProto((WWWContext, char *));

#endif
