/*
 * frame.c
 *
 * Copyright (c) 1995, John Kilburg <john@cs.unlv.edu>
 * See the file COPYRIGHT for details.
 */
#include <stdio.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include <X11/IntrinsicP.h>
#include "WWWP.h"

#define STACK_LOAD 0
#define STACK_BACK 1
#define STACK_RELOAD 2
#define STACK_DOWNLOAD 3
#define STACK_REPLACE 4

/*
 * Private functions
 */
static RenderHooks *GetRenderHooks _ArgProto((WWWContext, char *));
static ProtoHooks *GetProtoHooks _ArgProto((WWWContext, char *));
static Stack *StackCreate _ArgProto((Frame, URLParts *, int));
static void StackDestroy _ArgProto((Stack *));
static void StackProtoDestroy _ArgProto((Stack *));
static void StackClean _ArgProto((Stack *));
static void FrameStackPop _ArgProto((Frame));

/*
 * GetRenderHooks
 */
static RenderHooks *
GetRenderHooks(wc, content)
struct _WWWContext *wc;
char *content;
{
  RenderHooks *rh;
  size_t rlen;

  if (content == NULL) return(NULL);
  rlen = strlen(content);
  for (rh = wc->rhlist; rh != NULL; rh = rh->next)
  {
    if (strlen(rh->content) == rlen && strcasecmp(rh->content, content) == 0)
    {
      return(rh);
    }
  }

  if (wc->mod_resolve != NULL)
  {
    if ((wc->mod_resolve)(wc, wc->mod_closure, WWW_RENDER, content))
    {
      for (rh = wc->rhlist; rh != NULL; rh = rh->next)
      {
	if (strlen(rh->content) == rlen &&
	    strcasecmp(rh->content, content) == 0)
	{
	  return(rh);
	}
      }
    }
  }

  return(NULL);
}

/*
 * GetProtoHooks
 */
static ProtoHooks *
GetProtoHooks(wc, proto)
struct _WWWContext *wc;
char *proto;
{
  ProtoHooks *ph;
  size_t plen;

  if (proto == NULL) return(NULL);
  plen = strlen(proto);
  for (ph = wc->phlist; ph != NULL; ph = ph->next)
  {
    if (strlen(ph->protocol) == plen && strcasecmp(ph->protocol, proto) == 0)
    {
      return(ph);
    }
  }

  if (wc->mod_resolve != NULL)
  {
    if ((wc->mod_resolve)(wc, wc->mod_closure, WWW_PROTO, proto))
    {
      for (ph = wc->phlist; ph != NULL; ph = ph->next)
      {
	if (strlen(ph->protocol) == plen &&
	    strcasecmp(ph->protocol, proto) == 0)
	{
	  return(ph);
	}
      }
    }
  }

  return(NULL);
}

/*
 * StackProtoDestroy
 */
static void
StackProtoDestroy(sp)
Stack *sp;
{
  if (sp->ph != NULL && sp->ph->destroy != NULL && sp->pc != NULL)
  {
    (sp->ph->destroy)(sp->pc);
  }
  sp->pc = NULL;
  return;
}

/*
 * StackClean
 */
static void
StackClean(sp)
Stack *sp;
{
  if (sp->rh != NULL && sp->rh->destroy != NULL && sp->rc != NULL)
  {
    (sp->rh->destroy)(sp->rc);    
  }
  sp->rc = NULL;

  StackProtoDestroy(sp);

  if (sp->crc != NULL && sp->crh->destroy != NULL && sp->crc != NULL)
  {
    (sp->crh->destroy)(sp->crc);
  }
  sp->crc = NULL;

  return;
}

/*
 * StackDestroy
 */
void
StackDestroy(sp)
Stack *sp;
{
  StackClean(sp);
  if (sp->d != NULL) DocumentDestroy(sp->d);
  MPDestroy(sp->mp);

  return;
}

/*
 * StackCreate
 */
static Stack *
StackCreate(frame, up, action)
struct _Frame *frame;
URLParts *up;
int action;
{
  Stack *sp;
  ProtoHooks *ph;
  MemPool mp;

  if (!URLIsAbsolute(up)) return(NULL);

  mp = MPCreate();
  sp = (Stack *)MPCGet(mp, sizeof(Stack));
  sp->mp = mp;
  sp->action = action;
  sp->frame = frame;
  sp->up = URLDup(sp->mp, up);

  /*
   * Look in the cache.
   */
  if (action != STACK_RELOAD)
  {
    if ((ph = GetProtoHooks(frame->wc, "cache-read")) != NULL)
    {
      if ((sp->pc = (ph->init)(sp->up, NULL, ph->class_context, frame))!=NULL)
      {
	sp->ph = ph;
	return(sp);
      }
    }
  }
  else sp->up->reload = true;

  /*
   * Check for a proxy.
   */
  if ((sp->pup = GetProxy(sp->mp, up)) != NULL)
  {
    if ((ph = GetProtoHooks(frame->wc, sp->pup->protocol)) != NULL)
    {
      if ((sp->pc = (ph->init)(sp->pup, sp->up,
			       ph->class_context, frame)) != NULL)
      {
	sp->ph = ph;
	return(sp);
      }
    }
    StackDestroy(sp);
    return(NULL);
  }

  /*
   * Try to get it directly.
   */
  if ((ph = GetProtoHooks(frame->wc, sp->up->protocol)) != NULL)
  {
    if ((sp->pc = (ph->init)(sp->up, NULL, ph->class_context, frame)) != NULL)
    {
      sp->ph = ph;
      return(sp);
    }
    StackDestroy(sp);
    return(NULL);
  }

  StackDestroy(sp);

  return(NULL);
}

/*
 * FrameStackPop
 */
static void
FrameStackPop(frame)
struct _Frame *frame;
{
  Stack *t;

  if (frame->slist == NULL) return;

  t = frame->slist;
  frame->slist = frame->slist->next;
  StackDestroy(t);
  return;
}

/*
 * Public functions
 */

/*
 * FrameEndData
 */
int
FrameEndData(frame, d)
struct _Frame *frame;
Document *d;
{
  Stack *sp = frame->sp;

  sp->d = d;
  if (sp->rh != NULL && sp->rh->end != NULL && sp->rc != NULL)
  {
    (sp->rh->end)(sp->rc, d);
  }
  if (sp->crh != NULL && sp->crh->add != NULL && sp->crc != NULL)
  {
    (sp->crh->end)(sp->crc, d);
  }

  return(0);
}

/*
 * FrameAddData
 */
int
FrameAddData(frame, d)
struct _Frame *frame;
Document *d;
{
  Stack *sp = frame->sp;

  sp->d = d;
  if (sp->rh != NULL && sp->rh->add != NULL && sp->rc != NULL)
  {
    (sp->rh->add)(sp->rc, d);
  }
  if (sp->crh != NULL && sp->crh->add != NULL && sp->crc != NULL)
  {
    (sp->crh->add)(sp->crc, d);
  }

  return(0);
}

/*
 * FrameInitData
 */
int
FrameInitData(frame, d)
struct _Frame *frame;
Document *d;
{
  Stack *sp = frame->sp;
  char *value;
  char *url;

  sp->d = d;

  if (sp->action == STACK_DOWNLOAD) value = "x-www/x-download";
  else value = MIMEGetField(d->mflist, "content-type");
  if (value != NULL)
  {  
    if ((sp->rh = GetRenderHooks(frame->wc, value)) != NULL)
    {
      sp->rc = (sp->rh->init)(sp->up, d, sp->rh->class_context, frame);
      if (sp->rc != NULL)
      {
	if (frame->www != NULL) WWWDrawReset(frame->www);
	if (sp->action == STACK_BACK)
	{
	  FrameStackPop(frame);
	  FrameStackPop(frame);
	}
	else if (sp->action == STACK_RELOAD || sp->action == STACK_REPLACE)
	{
	  FrameStackPop(frame);
	}
	if (sp != frame->slist)
	{
	  if (frame->slist != NULL) StackClean(frame->slist);
	  sp->next = frame->slist;
	  frame->slist = sp;
	}
	if (frame->url_callback != NULL)
	{
	  url = URLMakeString(sp->mp, frame->slist->up, false);
	  (frame->url_callback)(frame, frame->url_closure, url);
	}
      }
    }
  }

  /*
   * This is needed so that the parent window will know that the
   * child can't be rendered.
   */
  if (sp->rc == NULL)
  {
    if (frame->size_callback != NULL)
    {
      (frame->size_callback)(frame->size_closure, -1, 0, 0);
    }
  }

  /*
   * Get the cache "renderer"
   */
  if ((sp->crh = GetRenderHooks(frame->wc,
				"x-www/x-cache-write")) != NULL)
  {
    sp->crc = (sp->crh->init)(sp->up, d, sp->crh->class_context, frame);
  }

  return(0);
}

/*
 * FrameCancel
 */
int
FrameCancel(frame)
struct _Frame *frame;
{
  Stack *sp = frame->sp;
  struct _Frame *c;

  if (frame->destroyed) return(0);

  if (sp == NULL) return(0);

  for (c = frame->kids; c != NULL; c = c->knext)
  {
    FrameCancel(c);
  }

  if (frame->sp == frame->slist) StackProtoDestroy(sp);
  else StackDestroy(sp);
  frame->sp = NULL;

  return(0);
}

/*
 * FrameCreate
 */
Frame
FrameCreate(wc, parent)
struct _WWWContext *wc;
struct _Frame *parent;
{
  struct _Frame *frame;
  MemPool mp;

  mp = MPCreate();
  frame = (struct _Frame *)MPCGet(mp, sizeof(struct _Frame));
  frame->mp = mp;

  if (parent != NULL)
  {
    frame->toplevel = parent->toplevel;
    frame->parent = parent;
    frame->knext = parent->kids;
    parent->kids = frame;
  }

  frame->wc = wc;
  frame->fnext = wc->flist;
  wc->flist = frame;

  return(frame);
}

/*
 * FrameAssignName
 */
void
FrameAssignName(frame, name)
struct _Frame *frame;
char *name;
{
  frame->name = MPStrDup(frame->mp, name);
  return;
}

/*
 * FrameCreateInv
 */
Frame
FrameCreateInv(parent)
struct _Frame *parent;
{
  return(FrameCreate(parent->wc, parent));
}

/*
 * FrameDestroy
 */
int
FrameDestroy(frame)
struct _Frame *frame;
{
  Stack *c, *t;

  if (frame->destroyed) return(0);
  frame->destroyed = true;

  if (frame->sp != NULL && frame->sp != frame->slist)
  {
    StackDestroy(frame->sp);
  }

  for (c = frame->slist; c != NULL; )
  {
    t = c;
    c = c->next;
    StackDestroy(t);
  }

  frame->slist = NULL;
  frame->sp = NULL;
  if (frame->www != NULL) XtDestroyWidget(frame->www);

  return(0);
}

/*
 * FrameLoad
 */
int
FrameLoad(frame, up)
struct _Frame *frame;
URLParts *up;
{
  Stack *sp;

  if ((sp = StackCreate(frame, up, STACK_LOAD)) != NULL)
  {
    FrameCancel(frame);
    frame->sp = sp;
  }

  return(0);
}

/*
 * FrameHome
 */
int
FrameHome(frame)
struct _Frame *frame;
{
  Stack *c, *t;
  URLParts *up = NULL;
  MemPool mp;

  if (frame->sp != NULL && frame->slist->next != NULL)
  {
    mp = MPCreate();
    StackDestroy(frame->sp);
    
    for (c = frame->slist; c != NULL; )
    {
      t = c;
      c = c->next;
      if (c == NULL) up = URLDup(mp, t->up);
      StackDestroy(t);
    }
    
    frame->slist = NULL;
    frame->sp = NULL;
    
    if (up != NULL) FrameLoad(frame, up);
    MPDestroy(mp);
  }

  return(0);
}

/*
 * FrameReload
 */
int
FrameReload(frame)
struct _Frame *frame;
{
  Stack *sp;

  if (frame->slist != NULL)
  {
    if ((sp = StackCreate(frame, frame->slist->up, STACK_RELOAD)) != NULL)
    {
      FrameCancel(frame);
      frame->sp = sp;
    }
  }

  return(0);
}

/*
 * FrameDownload
 */
int
FrameDownload(frame, up)
struct _Frame *frame;
URLParts *up;
{
  Stack *sp;

  if ((sp = StackCreate(frame, up, STACK_DOWNLOAD)) != NULL)
  {
    FrameCancel(frame);
    frame->sp = sp;
  }

  return(0);
}

/*
 * FrameBack
 */
int
FrameBack(frame)
struct _Frame *frame;
{
  Stack *sp;

  if (frame->slist != NULL && frame->slist->next != NULL)
  {
    if ((sp = StackCreate(frame, frame->slist->next->up, STACK_BACK)) != NULL)
    {
      FrameCancel(frame);
      frame->sp = sp;
    }
  }

  return(0);
}

/*
 * FrameReplace
 */
int
FrameReplace(frame, up)
struct _Frame *frame;
URLParts *up;
{
  Stack *sp;

  if ((sp = StackCreate(frame, up, STACK_REPLACE)) != NULL)
  {
    FrameCancel(frame);
    frame->sp = sp;
  }

  return(0);
}

/*
 * FrameOverride
 */
int
FrameOverride(frame, orh, orc)
struct _Frame *frame;
RenderHooks *orh;
void *orc;
{
  frame->use_override = true;
  memcpy(&(frame->orh), orh, sizeof(RenderHooks));
  frame->orc = orc;
  return(0);
}

/*
 *
 * Document handling convienence stuff
 *
 */

/*
 * DocumentCreate
 */
Document *
DocumentCreate(nocache)
bool nocache;
{
  Document *d;
  MemPool mp;

  mp = MPCreate();
  d = (Document *)MPCGet(mp, sizeof(Document));
  d->nocache = nocache;
  d->mp = mp;

  return(d);
}

/*
 * DocumentDestroy
 */
void
DocumentDestroy(d)
Document *d;
{
  if (d->ds != NULL) free_mem(d->ds);
  MPDestroy(d->mp);

  return;
}
