/*
 * io.c
 *
 * Copyright (C) 1994, 1995, John Kilburg <john@cs.unlv.edu>
 * See the file COPYRIGHT for details.
 */
#include <stdio.h>
#include <time.h>
#include <errno.h>
#include <fcntl.h>
#if ((defined(SYSV) || defined(SVR4)) && defined(sun))
#include <sys/file.h>
#endif
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <X11/IntrinsicP.h>

#include "WWWP.h"

#define READ_QUANTUM 256
#define ALLOC_AHEAD 10    /* minimum number of BUFSIZ blocks to alloc when
			     memory is needed for the next read.
			     has to be >= 1
                           */

struct _IOStream
{
  MemPool mp;
  struct _Frame *frame;
  int s;
  void (*efunc)();
  void *closure;
  XtAppContext appcon;
  XtInputId xid;

  /* message stuff */
  char *mbuf;
  size_t mlen;
  char *wstr;
  char *rstr;

  /* read stuff */
  void (*rdfunc)();
  byte *rdb;
  size_t rdlen;
  size_t rdmax;
  size_t rdbsize;

  /* write stuff */
  void (*wrfunc)();
  size_t wri;
  byte *wrb;
  size_t wrmax;
};

struct _IOTimeOut
{
  MemPool mp;
  XtAppContext appcon;
  XtIntervalId iid;
  void (*func)();
  void *closure;
};

static void RemoveHandlers _ArgProto((struct _IOStream *));
static void AddHandlers _ArgProto((struct _IOStream *,
				   XtInputMask, void (*)()));
static char *MakeMessage _ArgProto((struct _IOStream *, char *, size_t));

/*
 * MakeMessage
 */
static char *
MakeMessage(ps, str, n)
struct _IOStream *ps;
char *str;
size_t n;
{
  size_t len;

  len = strlen(str) + 100 + 100 + 1;
  if (len > ps->mlen)
  {
    ps->mbuf = (char *)MPGet(ps->mp, len + 1);
    ps->mlen = len;
  }
  sprintf (ps->mbuf, str, ps->frame->wc->iocnt, n);

  return(ps->mbuf);
}

/*
 * RemoveHandlers
 */
static void
RemoveHandlers(ps)
struct _IOStream *ps;
{
  if (ps->xid != 0)
  {
    XtRemoveInput(ps->xid);
    ps->xid = 0;
  }
  return;
}

/*
 * AddHandlers
 */
static void
AddHandlers(ps, iomask, iohandler)
struct _IOStream *ps;
XtInputMask iomask;
void (*iohandler)();
{
  ps->xid = XtAppAddInput(ps->appcon,
			  ps->s, (XtPointer)iomask,
			  iohandler, (XtPointer)ps);
  return;
}

/*
 * WriteStreamHandler
 */
static void
WriteStreamHandler(cldata, netfd, xid)
XtPointer cldata;
int *netfd;
XtInputId *xid;
{
  int rval;
  struct _IOStream *ps = (struct _IOStream *)cldata;

  rval = write(ps->s, ps->wrb + ps->wri, ps->wrmax - ps->wri);
  if (rval < 0)
  {
    if (errno != EWOULDBLOCK)
    {
      RemoveHandlers(ps);
      if (ps->efunc != NULL) (ps->efunc)(ps->closure);
    }
  }
  else
  {
    ps->wri += (size_t)rval;
    FrameMessage(ps->frame, MakeMessage(ps, ps->wstr, ps->wri));
    if (ps->wri == ps->wrmax)
    {
      RemoveHandlers(ps);
      (ps->wrfunc)(ps, ps->closure);
    }
  }

  return;
}

/*
 * ReadStreamHandler
 */
static void
ReadStreamHandler(cldata, netfd, xid)
XtPointer cldata;
int *netfd;
XtInputId *xid;
{
  struct _IOStream *ps = (struct _IOStream *)cldata;
  int rlen;
  size_t readsize;

  /*
   * Figure out how much to read.  Try for READ_QUANTUM unless there isn't
   * that much left to read.
   */
  if (ps->rdmax > 0)
  {
    size_t t = ps->rdmax - ps->rdlen;
    if (READ_QUANTUM < t) readsize = READ_QUANTUM;
    else readsize = t;
  }
  else readsize = READ_QUANTUM;

  /*
   * Allocate buffer space for the read.  If there has been no allocation
   * then allocate the maximum if we know it so we don't have to allocate
   * memory again.  If we have already allocated some memory then realloc
   * more memory to accomodate the new read.
   */
  if (ps->rdb == NULL)
  {
    ps->rdbsize = (ps->rdmax / BUFSIZ + 1) * BUFSIZ;
    ps->rdb = (byte *)alloc_mem(ps->rdbsize);
  }
  if (readsize > ps->rdbsize - ps->rdlen)
  {
    ps->rdbsize += (readsize / BUFSIZ + ALLOC_AHEAD) * BUFSIZ;
    ps->rdb = (byte *)realloc_mem(ps->rdb, ps->rdbsize);
  }

  rlen = read(ps->s, ps->rdb + ps->rdlen, readsize);
  if (rlen < 0)
  {
    if (errno != EWOULDBLOCK)
    {
      RemoveHandlers(ps);
      if (ps->efunc != NULL) (ps->efunc)(ps->closure);
    }
  }
  else if (rlen > 0)
  {
    ps->rdlen += (size_t)rlen;
    FrameMessage(ps->frame, MakeMessage(ps, ps->rstr,
					abs(ps->rdmax - ps->rdlen)));
    RemoveHandlers(ps);
    (ps->rdfunc)(ps, ps->rdmax > 0 && ps->rdlen == ps->rdmax, ps->closure);
  }
  else
  {
    RemoveHandlers(ps);
    (ps->rdfunc)(ps, true, ps->closure);
  }

  return;
}

/*
 * IOSetStreamBuffer
 */
void
IOSetStreamBuffer(ps, b, blen)
struct _IOStream *ps;
byte *b;
size_t blen;
{
  /* this should never happen */
  if (blen < ps->rdlen || b == NULL)
  {
    fprintf (stderr, "IOSetStreamBuffer: buffer NULL or too small\n");
    fflush(stderr);
    abort();
  }

  if (ps->rdb != NULL) free_mem(ps->rdb);
  ps->rdb = b;
  ps->rdmax = blen;

  if (blen > 0)
  {
    ps->rstr = FrameGetResource(ps->frame, "io.read_known");
    if (ps->rstr == NULL) ps->rstr = "io.read_known";
  }

  return;
}

/*
 * IOResetStream
 */
void
IOResetStream(ps)
struct _IOStream *ps;
{
  if (ps->rdb != NULL) free_mem(ps->rdb);
  ps->rdb = NULL;
  ps->rdmax = 0;
  ps->rdlen = 0;

  return;
}

/*
 * IOReadStream
 */
void
IOReadStream(ps, func)
struct _IOStream *ps;
void (*func)();
{
  ps->rdfunc = func;
  AddHandlers(ps, XtInputReadMask, ReadStreamHandler);

  return;
}

/*
 * IOWriteStream
 */
void
IOWriteStream(ps, b, blen, func)
struct _IOStream *ps;
byte *b;
size_t blen;
void (*func)();
{
  ps->wrb = b;
  ps->wrmax = blen;
  ps->wrfunc = func;
  ps->wri = 0;
  AddHandlers(ps, XtInputWriteMask, WriteStreamHandler);

  return;
}

/*
 * IOCreateStream
 */
IOStream
IOCreateStream(frame, host, port, efunc, closure)
struct _Frame *frame;
char *host;
int port;
void (*efunc)();
void *closure;
{
  struct _IOStream *ps;
  int s;
  int rval;
  struct sockaddr_in addr;
  struct hostent *hp;
  char *dbstr;
  MemPool mp;

  memset(&addr, 0, sizeof(addr));

  /* fix by Jim Rees so that numeric addresses are dealt with */
  if ((addr.sin_addr.s_addr = inet_addr(host)) == -1)
  {
    if ((hp = (struct hostent *)gethostbyname(host)) == NULL)
    {
      return(NULL);
    }
    memcpy(&(addr.sin_addr), hp->h_addr, hp->h_length);
  }

  addr.sin_family = AF_INET;
  addr.sin_port = htons((unsigned short)port);

  s = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (s < 0) return(NULL);

  fcntl(s, F_SETFL, FNDELAY);

  if ((rval = connect(s, (struct sockaddr *)&addr, sizeof(addr))) != 0)
  {
    if (errno != EINPROGRESS) return(NULL);
  }

  mp = MPCreate();
  ps = (struct _IOStream *)MPCGet(mp, sizeof(struct _IOStream));
  ps->mp = mp;
  ps->frame = frame;
  ps->s = s;
  ps->efunc = efunc;
  ps->closure = closure;
  ps->appcon = XtWidgetToApplicationContext(frame->www);

  ps->wstr = FrameGetResource(frame, "io.write");
  if (ps->wstr == NULL) ps->wstr = "io.write";
  ps->rstr = FrameGetResource(frame, "io.read_unknown");
  if (ps->rstr == NULL) ps->rstr = "io.read_unknown";

  dbstr = FrameGetResource(frame, "io.open");
  if (dbstr == NULL) dbstr = "io.open";

  /* make the message buffer fairly large so another is probably not needed */
  ps->mlen = strlen(dbstr) + strlen(host) + 100 + BUFSIZ;
  ps->mbuf = (char *)MPGet(mp, ps->mlen + 1);
  sprintf (ps->mbuf, dbstr, host, port);
  FrameMessage(frame, ps->mbuf);

  frame->wc->iocnt++;

  return((IOStream)ps);
}

/*
 * IODestroyStream
 */
void
IODestroyStream(ps)
struct _IOStream *ps;
{
  ps->frame->wc->iocnt--;
  RemoveHandlers(ps);
  close(ps->s);
  MPDestroy(ps->mp);

  return;
}

/*
 * IOStreamBuffer
 */
byte *
IOStreamBuffer(ps)
struct _IOStream *ps;
{
  return(ps->rdb);
}

/*
 * IOStreamBufferLength
 */
size_t
IOStreamBufferLength(ps)
struct _IOStream *ps;
{
  return(ps->rdlen);
}

/*
 * IOSnarfStreamBuffer
 */
byte *
IOSnarfStreamBuffer(ps)
struct _IOStream *ps;
{
  byte *buffer = ps->rdb;
  ps->rdb = NULL;
  return(buffer);
}

/*
 * IOCreateTimeOut
 */  
IOTimeOut
IOCreateTimeOut(frame, closure)
struct _Frame *frame;
void *closure;
{
  struct _IOTimeOut *pto;
  MemPool mp;

  mp = MPCreate();
  pto = (struct _IOTimeOut *)MPCGet(mp, sizeof(struct _IOTimeOut));
  pto->mp = mp;
  pto->appcon = XtWidgetToApplicationContext(frame->www);
  pto->closure = closure;

  return(pto);
}

/*
 * TimeOutHandler
 */
static void
TimeOutHandler(closure, iid)
XtPointer closure;
XtIntervalId iid;
{
  struct _IOTimeOut *pto = (struct _IOTimeOut *)closure;

  pto->iid = 0;
  (pto->func)((struct _IOTimeOut *)closure, pto->closure);

  return;
}

/*
 * IOSetTimeOut
 */
void
IOSetTimeOut(pto, interval, func)
struct _IOTimeOut *pto;
unsigned int interval;
void (*func)();
{
  pto->func = func;
  pto->iid = XtAppAddTimeOut(pto->appcon,
			     (unsigned long)interval,
			     TimeOutHandler,
			     pto);
  
  return;
}

/*
 * IODestroyTimeOut
 */
void
IODestroyTimeOut(pto)
struct _IOTimeOut *pto;
{
  if (pto->iid > 0) XtRemoveTimeOut(pto->iid);
  MPDestroy(pto->mp);
  return;
}
