/*
 * module.h
 *
 * Copyright (c) 1995, John D. Kilburg <john@cs.unlv.edu>
 * See the file COPYRIGHT for details.
 */

#ifndef __MODULE_H__
#define __MODULE_H__ 1

typedef void *IOStream;
typedef void *IOTimeOut;
typedef void *Frame;

typedef struct _render_hooks
{
  char *content;
  void *class_context;
  void *(*init)();
  int (*reset)();
  int (*add)();
  int (*end)();
  int (*destroy)();
  int (*mouse)();
  int (*keyboard)();
  int (*expose)();
  int (*motion)();
  int (*class_destroy)();
  struct _render_hooks *next;
} RenderHooks;

typedef struct _proto_hooks
{
  char *protocol;
  void *class_context;
  void *(*init)();
  int (*destroy)();
  int (*class_destroy)();
  struct _proto_hooks *next;
} ProtoHooks;

typedef struct _conv_hooks
{
  char *stype;
  char *dtype;
  void *class_context;
  void *(*init)();
  int (*add)();
  int (*end)();
  int (*destroy)();
  int (*class_destroy)();
  struct _conv_hooks *next;
} ConvHooks;

/*
 * Some of these fields should not be here.  It was convienent.
 *
 * protocol://username:password@hostname:port/filename#anchor
 */
typedef struct _url_parts
{
  char *protocol;
  char *hostname;
  int port;
  char *filename;
  char *anchor;
  char *method;
  void *input_data;
  size_t input_data_len;
  char *input_data_type;
  char *auth_type;
  char *auth_realm;
  char *username;
  char *password;
  bool reload;
} URLParts;

typedef struct _mimefield
{
  char *name;
  char *value;
  struct _mimefield *next;
} MIMEField;

typedef struct _document
{
  byte *ds;                        /* start of buffer, free this */
  byte *data;                      /* start of data */
  size_t len;                      /* length of data */
  MIMEField *mflist;               /* pseudo MIME stuff */
  bool nocache;                    /* don't cache document ? */
} Document;

typedef struct _mimecap
{
  char *ctype;
  char *command;
  struct _mimecap *next;  
} MIMECap;

/*
 * mime.c
 */
MIMEField *MIMECreateField _ArgProto((char *, char *));
void MIMEDestroyField _ArgProto((MIMEField *));
MIMEField *MIMEParseField _ArgProto((char *, size_t));
MIMEField *MIMEParseHeader _ArgProto((char *, size_t));

char *MIMEGetField _ArgProto((MIMEField *, char *));

char *MIMEExt2Content _ArgProto((char *));
void MIMEReadTypeFiles _ArgProto((char *));
void MIMEReadCapFiles _ArgProto((char *));
int MIMEStartViewer _ArgProto((MIMECap *, char *));
MIMECap *MIMEFindViewer _ArgProto((char *));
int MIMEFindData _ArgProto((char *, size_t));

/*
 * pipe.c
 */
int PipeCommand _ArgProto((char *, int *, int));

/*
 * io.c
 */
IOStream IOCreateStream _ArgProto((Frame, char *, int, void (*)(), void *));
void IODestroyStream _ArgProto((IOStream));
void IOWriteStream _ArgProto((IOStream, byte *, size_t, void (*)()));
void IOReadStream _ArgProto((IOStream, void (*)()));
void IOSetStreamBuffer _ArgProto((IOStream, byte *, size_t));
byte *IOStreamBuffer _ArgProto((IOStream));
int IOStreamBufferLength _ArgProto((IOStream));
byte *IOSnarfStreamBuffer _ArgProto((IOStream));
IOTimeOut IOCreateTimeOut _ArgProto((Frame, void *));
void IOSetTimeOut _ArgProto((IOTimeOut, unsigned int, void (*)()));
void IODestroyTimeOut _ArgProto((IOTimeOut));
void IOResetStream _ArgProto((IOStream));

/*
 * url.c
 */
int URLcmp _ArgProto((URLParts *, URLParts *));
char *URLMakeString _ArgProto((URLParts *, int));
URLParts *URLResolve _ArgProto((URLParts *, URLParts *));
URLParts *URLParse _ArgProto((char *));
void URLDestroy _ArgProto((URLParts *));
char *URLUnescape _ArgProto((char *));
char *URLEscape _ArgProto((char *, int));
URLParts *URLDup _ArgProto((URLParts *));
int URLIsAbsolute _ArgProto((URLParts *));
char *URLHash _ArgProto((URLParts *));

/*
 * frame.c
 */
int FrameInitData _ArgProto((Frame, Document *));
int FrameAddData _ArgProto((Frame, Document *));
int FrameEndData _ArgProto((Frame, Document *));

int FrameLoad _ArgProto((Frame, URLParts *));
int FrameReload _ArgProto((Frame));
int FrameDownload _ArgProto((Frame, URLParts *));
int FrameBack _ArgProto((Frame));
int FrameCancel _ArgProto((Frame));
int FrameHome _ArgProto((Frame));
int FrameClear _ArgProto((Frame));

Document *DocumentCreate _ArgProto((bool));
void DocumentDestroy _ArgProto((Document *));

/*
 * gui.c
 */
#define FrameSizeFixed 0
#define FrameSizeVariable 1

bool FrameOnScreen _ArgProto((Frame, int, int,
			      unsigned int, unsigned int));

int FrameGetNamedColor _ArgProto((Frame, char *, unsigned long *));

char *FrameGetResource _ArgProto((Frame, char *));

Frame FrameCreate _ArgProto((Frame, URLParts *, void (*)(), void *));
Frame FrameCreateSized _ArgProto((Frame, URLParts *,
				  unsigned int, unsigned int));
int FrameMap _ArgProto((Frame, int, int));
int FrameGetSize _ArgProto((Frame, int, unsigned int *, unsigned int *));
int FrameSetSize _ArgProto((Frame, int, unsigned int, unsigned int));
int FrameReturnSize _ArgProto((Frame, int, unsigned int, unsigned int));
int FrameDestroy _ArgProto((Frame));

int FrameGetAuth _ArgProto((Frame, void (*)(), void *));

int FrameMessage _ArgProto((Frame, char *));
void *FrameWidget _ArgProto((Frame));

int FrameBusy _ArgProto((Frame));
int FrameIdle _ArgProto((Frame));

/*
 * WWW.c
 */
int WWWAddProtoHooks _ArgProto((ProtoHooks *));
int WWWAddRenderHooks _ArgProto((RenderHooks *));
int WWWAddConvHooks _ArgProto((ConvHooks *));
char *WWWGetResource _ArgProto((char *));

#endif
