/*
 * proxy.c
 *
 * Copyright (c) 1995 Erik Corry ehcorry@inet.uni-c.dk
 * See the file COPYRIGHT for details
 */
#include <stdio.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_UNAME
#include <sys/utsname.h>
#endif

#include <X11/IntrinsicP.h>

#include "WWWP.h"

static bool match_no_proxy _ArgProto((char *, char *));
static bool match_no_proxy_list _ArgProto((char *, char *, char *, char *));
static bool match_host_name _ArgProto((char *, char *));

static bool
match_no_proxy(remote_host, no_proxy_domain)
char *remote_host;
char *no_proxy_domain;
{
  int remote_len = strlen(remote_host);
  int no_proxy_len = strlen(no_proxy_domain);
  int min_len = MIN(remote_len, no_proxy_len);

  /*
   * avoid matching parts of domain names not divided by dots.
   */
  if(remote_len < no_proxy_len)
    if(no_proxy_domain[no_proxy_len - min_len - 1] != '.')
      return false;
  if(remote_len > no_proxy_len)
    if(remote_host[remote_len - min_len - 1] != '.')
      return false;

  /*
   * Do ends of strings match?
   */
  if (strncmp(remote_host + remote_len - min_len,
              no_proxy_domain + no_proxy_len - min_len,
              min_len) == 0)
    return(true);
  else
    return(false);
}

/*
 * If no_proxy is not defined, then use the part that is common
 * between the proxy server's name and the uname of the current host.
 * I.e. if the proxy server is wwwproxy.firewall.dreadco.com, and the
 * local host is dilbert-sparc.widget-dept.dreadco.com, then the proxy
 * is not used for *.dreadco.com addresses. Or if your proxy is set to
 * http://wwwcache.hensa.ac.uk#8080/ and your machine is aiai.ed.ac.uk,
 * then ac.uk addresses will not go via the cache.
 */
static bool
match_no_proxy_list(remote_host, local_host, no_proxy_list, proxy_host)
char *remote_host;
char *local_host;
char *no_proxy_list;
char *proxy_host;
{
  if (no_proxy_list != NULL)
  {
    char *local_no_proxy_list;
    char *no_proxy_elt;
    /*
     * Step through comma-separated list of domains/machines for
     * which no proxy should be used
     */
    int i;
    int l = strlen(no_proxy_list);
    no_proxy_elt = local_no_proxy_list = (char *)alloc_mem(l + 1);
    for (i = 0; i <= l; i++)
    {
      char temp;
      if ((temp = local_no_proxy_list[i] = no_proxy_list[i]) == ',' ||
          temp == 0)
      {
        /* Zero out comma in copy of list */
        local_no_proxy_list[i] = 0;
        if(match_no_proxy(remote_host, no_proxy_elt))
        {
          free_mem(local_no_proxy_list);
          return true;
        }
        /* advance pointer to next element in copy of list */
        no_proxy_elt = local_no_proxy_list + i + 1;
      }
    }
    free_mem(local_no_proxy_list);
    return(false);
  }
  else
  {
    /*
     * Use default no_proxy. See comments above.
     */
    int i, j;
    int l, pl;
    if(!local_host) return false;
    l = strlen(local_host);
    pl = strlen(proxy_host);
    for (i = l - 1, j = pl -1; i && j; i--, j--)
    {
      if(local_host[i] != proxy_host[i])
      {
        /* We have found the first point where local_host and proxy_host
         * differ (searching back from end). Now find the first '.' in
         * the common suffix, since we don't want to match partial
         * names
         */
         for(; i < l; i++, j++)
           if(local_host[i] == '.') break;

         /* If the local_host and the proxy_host names don't have
          * an parts in common, then use the proxy, ie no match
          * with no_proxy, ie return false
          */
         if(local_host[i] == 0)
           return false;

         /* Now local_host + i points to default no_proxy value */
         return(match_no_proxy(remote_host, local_host+i));
      }

    }
    /* If we are here, then the entire string matched, Ie we are using
     * the local host as a proxy. In that case always use the proxy, ie
     * report no match, ie false
     */
    return(false);
  }
}

/*
 * Are they the same host? For example kroete2.freinet.de is the same as
 * kroete2. Doesn't match kgb.ussr with kgb.us
 */
static bool
match_host_name(remote_host, local_host)
char *remote_host;
char *local_host;
{
  int rl = strlen(remote_host);
  int ll = strlen(local_host);
  /* Same length, same string */
  if(ll == rl) return(strcmp(remote_host, local_host) == 0 ? true : false);
  if(ll < rl)
  {
    /*
     * kroete2. matches kroete2.freinet.de and
     * kroete2 matches kroete2.freinet.de
     */
    if(local_host[ll - 1] == '.' || remote_host[ll] == '.')
    {
      return(strncmp(remote_host, local_host, ll) == 0 ? true : false);
    }
    return(false);
  }
  else
  {
    /*
     * kroete2. matches kroete2.freinet.de and
     * kroete2 matches kroete2.freinet.de
     */
    if(remote_host[rl - 1] == '.' || local_host[rl] == '.')
    {
      return(strncmp(remote_host, local_host, rl) == 0 ? true : false);
    }
    return(false);
  }
}

/*
 * GetProxy
 *
 * There is no proxy for 'localhost'.
 * There is also no proxy for the local system name (as
 * returned by uname(2) or 'uname -n'). Otherwise do the
 * no_proxy checking.
 *
 * I scratched around here...so this is my fault. -john
 */
URLParts *
GetProxy(mp, up)
MemPool mp;
URLParts *up;
{
  URLParts *pup = NULL;
  char *pname;
  char *apl;
# ifdef HAVE_UNAME
  struct utsname thishostname;
# endif

  if (up->protocol == NULL) return(NULL);

  /*
   * Check for protocol specific proxy information.
   */
  pname = (char *)alloc_mem(strlen(up->protocol) + strlen("_proxy") + 1);
  strcpy(pname, up->protocol);
  strcat(pname, "_proxy");
  if (getenv(pname) != NULL) pup = URLParse(mp, getenv(pname));
  free_mem(pname);

  /*
   * Check for "all" proxy information.
   */
  if (pup == NULL)
  {
    /*
     * Check the all list to make sure that the protocol should be
     * used with "all".  If the list doesn't exist then just assume
     * "all" really means "all".
     */
    if ((apl = getenv("all_proxy_list")) != NULL)
    {
      pname = apl;
      while ((pname = mystrtok(apl, ',', &apl)) != NULL)
      {
	if (strlen(up->protocol) == strlen(pname) &&
	    strcmp(up->protocol, pname) == 0) break;
      }
      if (pname == NULL) return(NULL);
    }
    if (getenv("all_proxy") != NULL) pup = URLParse(mp, getenv("all_proxy"));
  }

  /*
   * No proxy information?
   */
  if (pup == NULL) return(NULL);

  /*
   * Found proxy information now make sure that it can be used.
   */
# ifdef HAVE_UNAME
  uname(&thishostname);
  if (up->hostname &&
      strcmp(up->hostname, "localhost") != 0
      && ! match_host_name(up->hostname, thishostname.nodename)
      && ! match_no_proxy_list(up->hostname,
			       thishostname.nodename,
			       getenv("no_proxy"),
			       pup->hostname))
  {
    return(pup);
  }
# else
  if (up->hostname &&
      strcmp(up->hostname, "localhost") != 0
      && ! match_no_proxy_list(up->hostname,
			       NULL,
			       getenv("no_proxy"),
			       pup->hostname))
  {
    return(pup);
  }
# endif

  return(NULL);
}
